# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from absl.testing import absltest
from absl.testing import parameterized

import jax
import jax.numpy as jnp

from nemos.third_party.jaxopt.jaxopt import GradientDescent
from nemos.third_party.jaxopt.jaxopt import objective
from nemos.third_party.jaxopt.jaxopt._src import test_util

from sklearn import datasets


N_CALLS = 0

class GradientDescentTest(test_util.JaxoptTestCase):

  def test_logreg_with_intercept(self):
    X, y = datasets.make_classification(n_samples=10, n_features=5, n_classes=3,
                                        n_informative=3, random_state=0)
    data = (X, y)
    l2reg = 100.0
    fun = objective.l2_multiclass_logreg_with_intercept
    n_classes = len(jnp.unique(y))

    W_init = jnp.zeros((X.shape[1], n_classes))
    b_init = jnp.zeros(n_classes)
    pytree_init = (W_init, b_init)
    gd = GradientDescent(fun=fun, tol=1e-3, maxiter=500)
    pytree_fit, info = gd.run(pytree_init, l2reg=l2reg, data=data)

    # Check optimality conditions.
    self.assertLessEqual(info.error, 5e-2)

    # Compare against sklearn.
    W_skl, b_skl = test_util.logreg_skl(X, y, l2reg, fit_intercept=True)
    self.assertArraysAllClose(pytree_fit[0], W_skl, atol=5e-2)
    self.assertArraysAllClose(pytree_fit[1], b_skl, atol=5e-2)

  def test_logreg_implicit_diff(self):
    X, y = datasets.load_digits(return_X_y=True)
    data = (X, y)
    l2reg = float(X.shape[0])
    fun = objective.l2_multiclass_logreg

    jac_num = test_util.logreg_skl_jac(X, y, l2reg)
    W_skl = test_util.logreg_skl(X, y, l2reg)

    # Make sure the decorator works.
    gd = GradientDescent(fun=fun, tol=1e-3, maxiter=10, implicit_diff=True)
    def wrapper(hyperparams):
      return gd.run(W_skl, hyperparams, data).params
    jac_custom = jax.jacrev(wrapper)(l2reg)
    self.assertArraysAllClose(jac_num, jac_custom, atol=1e-2)

  def test_logreg_implicit_diff_kwargs(self):
    # cf. https://github.com/google/jaxopt/issues/51
    X, y = datasets.load_digits(return_X_y=True)
    data = (X, y)
    l2reg = float(X.shape[0])
    fun = objective.l2_multiclass_logreg

    jac_num = test_util.logreg_skl_jac(X, y, l2reg)
    W_skl = test_util.logreg_skl(X, y, l2reg)

    # Make sure the decorator works.
    gd = GradientDescent(fun=fun, tol=1e-3, maxiter=10, implicit_diff=True)
    def wrapper(l2reg):
      return gd.run(W_skl, l2reg=l2reg, data=data).params
    jac_custom = jax.jacrev(wrapper)(l2reg)
    self.assertArraysAllClose(jac_num, jac_custom, atol=1e-2)

  @parameterized.product(acceleration=[True, False])
  def test_logreg_unrolled_autodiff(self, acceleration):
    X, y = datasets.load_digits(return_X_y=True)
    data = (X, y)
    l2reg = float(X.shape[0])
    fun = objective.l2_multiclass_logreg
    n_classes = len(jnp.unique(y))

    jac_l2reg = test_util.logreg_skl_jac(X, y, l2reg)

    # Compute the Jacobian w.r.t. l2reg via forward-mode autodiff.
    W_init = jnp.zeros((X.shape[1], n_classes))
    gd = GradientDescent(fun=fun, tol=1e-3, maxiter=200, implicit_diff=False,
                         acceleration=acceleration)
    def wrapper(hyperparams):
      return gd.run(W_init, hyperparams, data).params
    jac_l2reg2 = jax.jacfwd(wrapper)(l2reg)
    self.assertArraysAllClose(jac_l2reg, jac_l2reg2, atol=1e-2)

  def test_jit_and_vmap(self):
    X, y = datasets.make_classification(n_samples=30, n_features=5,
                                        n_informative=3, n_classes=2,
                                        random_state=0)
    data = (X, y)
    fun = objective.l2_multiclass_logreg
    hyperparams_list = jnp.array([1.0, 10.0])
    W_init = jnp.zeros((X.shape[1], 2))
    gd = GradientDescent(fun=fun, tol=1e-3, maxiter=100)

    def solve(hyperparams):
      W_fit, info = gd.run(W_init, hyperparams, data)
      return info.error

    errors = jnp.array([solve(hyperparams) for hyperparams in hyperparams_list])
    errors_vmap = jax.vmap(solve)(hyperparams_list)
    self.assertArraysAllClose(errors, errors_vmap, atol=1e-3)

    error0 = jax.jit(solve)(hyperparams_list[0])
    self.assertAllClose(errors[0], error0)

  @parameterized.product(n_iter=[10])
  def test_n_calls(self, n_iter):
    """Test whether the number of function calls
    is equal to the number of iterations + 1 in the
    no linesearch case, where the complexity is linear."""
    def fun(x):
      global N_CALLS
      N_CALLS += 1
      return sum(100.0*(x[1:] - x[:-1]**2.0)**2.0 + (1.0 - x[:-1])**2.0)

    x0 = jnp.zeros(2, dtype=jnp.float32)
    gd = GradientDescent(fun=fun, tol=1e-12, maxiter=n_iter, stepsize=1e-5, jit=False)
    gd.run(x0)
    self.assertEqual(N_CALLS, n_iter)

if __name__ == '__main__':
  # Uncomment the line below in order to run in float64.
  #jax.config.update("jax_enable_x64", True)
  absltest.main()



"""Collection of nemos typing."""

from __future__ import annotations

from typing import Any, Callable, NamedTuple, Tuple, TypeAlias, TypeVar, Union

import jax.numpy as jnp
import pynapple as nap
from jax.typing import ArrayLike
from numpy.typing import NDArray

from .pytrees import FeaturePytree

Pytree: TypeAlias = Any
Params: TypeAlias = Pytree
SolverState = TypeVar("SolverState")
StepResult: TypeAlias = Tuple[Params, SolverState]
DESIGN_INPUT_TYPE = Union[jnp.ndarray, FeaturePytree, nap.TsdFrame]

# copying jax.random's annotation
KeyArrayLike = ArrayLike

SolverRun = Callable[
    [
        Params,  # parameters, could be any pytree
        jnp.ndarray,  # Predictors (i.e. model design for GLM)
        jnp.ndarray,
    ],  # Output (neural activity)
    StepResult,
]

SolverInit = Callable[
    [
        Params,  # parameters, could be any pytree
        jnp.ndarray,  # Predictors (i.e. model design for GLM)
        jnp.ndarray,
    ],  # Output (neural activity)
    SolverState,
]

SolverUpdate = Callable[
    [
        Params,  # parameters, could be any pytree
        NamedTuple,
        jnp.ndarray,  # Predictors (i.e. model design for GLM)
        jnp.ndarray,
    ],  # Output (neural activity)
    StepResult,
]

ProximalOperator = Callable[
    [
        Params,  # parameters, could be any pytree
        float,  # Regularizer strength (for now float, eventually pytree)
        float,
    ],  # Step-size for optimization (must be a float)
    Tuple[jnp.ndarray, jnp.ndarray],
]

FeatureMatrix = nap.TsdFrame | NDArray

RegularizerStrength = float | Tuple[float, float]



"""Registry for mapping from solver name to concrete implementation."""

from typing import Type

from ._jaxopt_solvers import (
    JaxoptBFGS,
    JaxoptGradientDescent,
    JaxoptLBFGS,
    JaxoptNonlinearCG,
    JaxoptProximalGradient,
)
from ._svrg import WrappedProxSVRG, WrappedSVRG

solver_registry: dict[str, Type] = {
    "GradientDescent": JaxoptGradientDescent,
    #
    "ProximalGradient": JaxoptProximalGradient,
    #
    "LBFGS": JaxoptLBFGS,
    #
    "BFGS": JaxoptBFGS,
    #
    "SVRG": WrappedSVRG,
    "ProxSVRG": WrappedProxSVRG,
    #
    "NonlinearCG": JaxoptNonlinearCG,
}


def list_available_solvers():
    """
    List the available solvers that can be used for fitting models.

    To access an extended documentation about a specific solver,
    see `get_solver_documentation`.

    Example
    -------
    >>> import nemos as nmo
    >>> nmo.solvers.list_available_solvers()
    ['GradientDescent', 'ProximalGradient', 'LBFGS', 'BFGS', 'SVRG', 'ProxSVRG', 'NonlinearCG']
    >>> print(nmo.solvers.get_solver_documentation("SVRG"))
    Showing docstring of nemos.solvers._svrg.WrappedSVRG.
    For potentially more info, use `show_help=True`.
    <BLANKLINE>
    Adapter for NeMoS's implementation of SVRG following the AbstractSolver interface.
    <BLANKLINE>
    ...
    """
    return list(solver_registry.keys())



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from absl.testing import absltest
from absl.testing import parameterized
import jax
import jax.numpy as jnp
from jax import random
from jax.tree_util import tree_map
from jax.flatten_util import ravel_pytree
from nemos.third_party.jaxopt.jaxopt import BacktrackingLineSearch
from nemos.third_party.jaxopt.jaxopt import LBFGS
from nemos.third_party.jaxopt.jaxopt import GradientDescent
from nemos.third_party.jaxopt.jaxopt import objective
from nemos.third_party.jaxopt.jaxopt import OptStep
from nemos.third_party.jaxopt.jaxopt._src import test_util
from nemos.third_party.jaxopt.jaxopt._src.lbfgs import inv_hessian_product
from nemos.third_party.jaxopt.jaxopt._src.lbfgs import select_ith_tree
from nemos.third_party.jaxopt.jaxopt.tree_util import tree_sum
import numpy as onp
import scipy.optimize as scipy_opt
from sklearn import datasets


N_CALLS = 0

# Uncomment this line to test in x64 
# jax.config.update('jax_enable_x64', True)

def _materialize_inv_hessian(s_history, y_history, rho_history, start):
  history_size, n_dim = s_history.shape

  s_history = jnp.roll(s_history, -start, axis=0)
  y_history = jnp.roll(y_history, -start, axis=0)
  rho_history = jnp.roll(rho_history, -start, axis=0)

  I = jnp.eye(n_dim, n_dim)
  H = I

  for k in range(history_size):
    V = I - rho_history[k] * jnp.outer(y_history[k], s_history[k])
    H = (jnp.dot(V.T, jnp.dot(H, V)) +
         rho_history[k] * jnp.outer(s_history[k], s_history[k]))

  return H


def _inv_hessian_product(s, y, v, gamma=1.0):
  history_size = len(s)

  if history_size == 0:
    return v

  s = jnp.array(s)
  y = jnp.array(y)

  rho = jnp.zeros(history_size)
  alpha = jnp.zeros(history_size)

  # Compute right product.
  def body_fun(j, tup):
    rho, alpha, r = tup
    i = history_size - j - 1
    # rho[i] = 1. / jnp.sum(s[i] * y[i])
    rho = rho.at[i].set(1. / jnp.sum(s[i] * y[i]))
    # alpha[i] = rho[i] * jnp.sum(s[i] * r)
    alpha = alpha.at[i].set(rho[i] * jnp.sum(s[i] * r))
    r = r - alpha[i] * y[i]
    return rho, alpha, r

  # for i in reversed(range(history_size)):
  rho, alpha, r = jax.lax.fori_loop(0, history_size, body_fun, (rho, alpha, v))

  r = r * gamma

  # Compute left product.
  def body_fun2(i, r):
    beta = rho[i] * jnp.sum(y[i] * r)
    return r + s[i] * (alpha[i] - beta)

  # for i in range(history_size):
  r = jax.lax.fori_loop(0, history_size, body_fun2, r)

  return r


def _lbfgs(fun, init, stepsize=1e-3, maxiter=500, tol=1e-3,
           verbose=0, history_size=10, use_gamma=True):

  value_and_grad_fun = jax.value_and_grad(fun)

  x = init
  value, grad = value_and_grad_fun(init)
  s_history = []
  y_history = []

  for it in range(maxiter):
    if use_gamma and it > 0:
      gamma = jnp.vdot(y_history[-1], s_history[-1])
      gamma /= jnp.sum(y_history[-1] ** 2)
    else:
      gamma = 1.0

    direction = -_inv_hessian_product(s_history, y_history, grad, gamma)
    x_old, grad_old = x, grad
    x = x + stepsize * direction
    value, grad = value_and_grad_fun(x)

    s_history.append(x - x_old)
    y_history.append(grad - grad_old)

    if len(s_history) > history_size:
      s_history = s_history[1:]  # Pop left.
      y_history = y_history[1:]

    grad_norm = jnp.sqrt(jnp.sum(grad ** 2))

    if verbose:
      print("iter %d" % it, "f:", value, "grad:", grad_norm)

    if grad_norm <= tol:
      break

  return x


def get_fun(name, np):

  def rosenbrock(x):
    return np.sum(100. * np.diff(x) ** 2 + (1. - x[:-1]) ** 2)

  def himmelblau(p):
    x, y = p
    return (x ** 2 + y - 11.) ** 2 + (x + y ** 2 - 7.) ** 2

  def matyas(p):
    x, y = p
    return 0.26 * (x ** 2 + y ** 2) - 0.48 * x * y

  def eggholder(p):
    x, y = p
    return - (y + 47) * np.sin(np.sqrt(np.abs(x / 2. + y + 47.))) - x * np.sin(
      np.sqrt(np.abs(x - (y + 47.))))

  def zakharov(x):
    ii = np.arange(1, len(x) + 1, step=1, dtype=x.dtype)
    sum1 = (x**2).sum()
    sum2 = (0.5*ii*x).sum()
    answer = sum1+sum2**2+sum2**4
    return answer
  
  funs = dict(
    rosenbrock=rosenbrock,
    himmelblau=himmelblau,
    matyas=matyas,
    eggholder=eggholder,
    zakharov=zakharov
  )
  return funs[name]


class LbfgsTest(test_util.JaxoptTestCase):

  def test_inv_hessian_product_array(self):
    rng = onp.random.RandomState(0)
    history_size = 4
    d = 3
    s_history = jnp.array(rng.randn(history_size, d))
    y_history = jnp.array(rng.randn(history_size, d))
    rho_history = 1.0 / jnp.sum(s_history * y_history, axis=1)
    v = jnp.array(rng.randn(d))

    Hv1 = _inv_hessian_product(s_history, y_history, v)

    H = _materialize_inv_hessian(s_history, y_history, rho_history, 0)
    Hv2 = H.dot(v)

    Hv3 = inv_hessian_product(v, s_history, y_history, rho_history, start=0)

    self.assertArraysAllClose(Hv1, Hv2, atol=1e-2)
    self.assertArraysAllClose(Hv1, Hv3, atol=1e-2)

  @parameterized.product(start=[0, 1, 2, 3])
  def test_inv_hessian_product_pytree(self, start):
    rng = onp.random.RandomState(0)
    history_size = 4
    shape1 = (3, 2)
    shape2 = (5,)

    s_history1 = jnp.array(rng.randn(history_size, *shape1))
    s_history2 = jnp.array(rng.randn(history_size, *shape2))

    y_history1 = jnp.array(rng.randn(history_size, *shape1))
    y_history2 = jnp.array(rng.randn(history_size, *shape2))

    v1 = jnp.array(rng.randn(*shape1))
    v2 = jnp.array(rng.randn(*shape2))

    s_history = (s_history1, s_history2)
    y_history = (y_history1, y_history2)
    pytree = (v1, v2)

    v = ravel_pytree(pytree)[0]
    s = [ravel_pytree(select_ith_tree(s_history, i))[0]
          for i in range(history_size)]
    y = [ravel_pytree(select_ith_tree(y_history, i))[0]
          for i in range(history_size)]
    s, y = jnp.stack(s), jnp.stack(y)

    inv_rho_history = [jnp.vdot(s[i], y[i])
                       for i in range(history_size)]
    rho_history = 1./ jnp.array(inv_rho_history)

    Hv = inv_hessian_product(pytree, s_history, y_history, rho_history,
                             start=start)
    Hv = ravel_pytree(Hv)[0]

    H = _materialize_inv_hessian(s, y, rho_history, start)
    Hv_mat = jnp.dot(H, v)

    self.assertArraysAllClose(Hv, Hv_mat, atol=1e-5, rtol=1e-5)

  @parameterized.product(use_gamma=[True, False])
  def test_correctness(self, use_gamma):
    def fun(x, *args, **kwargs):  # Rosenbrock function.
      return sum(100.0*(x[1:] - x[:-1]**2.0)**2.0 + (1 - x[:-1])**2.0)

    x0 = jnp.zeros(2)
    lbfgs = LBFGS(fun=fun, tol=1e-3, stepsize=1e-3, maxiter=15, history_size=5,
                  use_gamma=use_gamma)
    x1, _ = lbfgs.run(x0)

    x2 = _lbfgs(fun, x0, stepsize=1e-3, maxiter=15, history_size=5,
                use_gamma=use_gamma)

    self.assertArraysAllClose(x1, x2, atol=1e-5)

  @parameterized.product(
      use_gamma=[True, False],
      linesearch=[
          "backtracking",
          "zoom",
          "hager-zhang",
          BacktrackingLineSearch(objective.binary_logreg, decrease_factor=0.5),
      ],
      linesearch_init=["max", "current", "increase"]
  )
  def test_binary_logreg(self, use_gamma, linesearch, linesearch_init):
    X, y = datasets.make_classification(
        n_samples=10, n_features=5, n_classes=2, n_informative=3, random_state=0
    )
    data = (X, y)
    fun = objective.binary_logreg

    w_init = jnp.zeros(X.shape[1])
    lbfgs = LBFGS(fun=fun, tol=1e-3, maxiter=500, use_gamma=use_gamma,
                  linesearch=linesearch, linesearch_init=linesearch_init)
    # Test with keyword argument.
    w_fit, info = lbfgs.run(w_init, data=data)

    # Check optimality conditions.
    self.assertLessEqual(info.error, 5e-2)

    # Compare against sklearn.
    w_skl = test_util.logreg_skl(X, y, 1e-6, fit_intercept=False,
                                 multiclass=False)
    self.assertArraysAllClose(w_fit, w_skl, atol=5e-2)

  @parameterized.product(
      use_gamma=[True, False],
      linesearch=[
          "backtracking",
          "zoom",
          "hager-zhang",
          BacktrackingLineSearch(
              objective.multiclass_logreg_with_intercept, decrease_factor=0.5
          ),
      ],
  )
  def test_multiclass_logreg(self, use_gamma, linesearch):
    data = datasets.make_classification(
        n_samples=10, n_features=5, n_classes=3, n_informative=3, random_state=0
    )
    fun = objective.multiclass_logreg_with_intercept

    w_init = jnp.zeros((data[0].shape[1], 3))
    b_init = jnp.zeros(3)
    pytree_init = (w_init, b_init)

    lbfgs = LBFGS(fun=fun, tol=1e-3, maxiter=500, use_gamma=use_gamma,
                  linesearch=linesearch)
    # Test with positional argument.
    _, info = lbfgs.run(pytree_init, data)

    # Check optimality conditions.
    self.assertLessEqual(info.error, 1e-2)

  @parameterized.product(implicit_diff=[True, False])
  def test_implicit_diff(self, implicit_diff): 
    fun = get_fun('rosenbrock', jnp)
    x0 = jnp.zeros(2)
    lbfgs = LBFGS(fun=fun, tol=1e-3, maxiter=500, implicit_diff=implicit_diff)
    x, _ = lbfgs.run(x0)

    # the Rosenbrock function is zero at its minimum
    self.assertLessEqual(fun(x), 1e-2)

  def test_warm_start_hessian_approx(self):
    def fun(x, *args, **kwargs):
      return sum(100.0*(x[1:] - x[:-1]**2.0)**2.0 + (1 - x[:-1])**2.0)

    x0 = jnp.zeros(2)
    lbfgs = LBFGS(fun=fun, tol=1e-3, maxiter=2, stepsize=1e-3)
    x1, lbfgs_state = lbfgs.run(x0)

    # warm start with the previous solution
    x2, _ = lbfgs.run(OptStep(x1, lbfgs_state))

  @parameterized.product(out_dtype=[jnp.float32, jnp.float64])
  def test_correct_dtypes(self, out_dtype):
    def f(x):
      return  jnp.cos(jnp.sum(jnp.exp(-x)) ** 2).astype(out_dtype)

    with jax.experimental.enable_x64():
      x0 = jnp.ones([5, 5], dtype=jnp.float32)
      lbfgs = LBFGS(fun=f, tol=1e-3, maxiter=500)
      x, state = lbfgs.run(x0)
      self.assertEqual(x.dtype, jnp.float32)
      for name in ("iter_num",):
        self.assertEqual(getattr(state, name).dtype, jnp.int64)
      for name in ("error", "s_history", "y_history", "rho_history"):
        self.assertEqual(getattr(state, name).dtype, jnp.float32, name)
      for name in ("value",):
        self.assertEqual(getattr(state, name).dtype, out_dtype, name)

  @parameterized.product(n_iter=[10])
  def test_n_calls(self, n_iter):
    """Test whether the number of function calls
    is equal to the number of iterations + 1 in the
    no linesearch case, where the complexity is linear."""
    global N_CALLS
    N_CALLS = 0

    def fun(x, *args, **kwargs):
      global N_CALLS
      N_CALLS += 1
      return sum(100.0*(x[1:] - x[:-1]**2.0)**2.0 + (1 - x[:-1])**2.0)

    x0 = jnp.zeros(2)
    lbfgs = LBFGS(fun=fun, tol=1e-7, maxiter=n_iter, stepsize=1e-3, jit=False)
    x, _ = lbfgs.run(x0)

    self.assertEqual(N_CALLS, n_iter + 1)

  def test_binary_logit_log_likelihood(self):
    # See issue #409
    rng = jax.random.PRNGKey(42)
    N = 1000
    beta = jnp.array([[0.5,0.5]]).T
    income = jax.random.normal(rng, shape=(N,1))
    x = jnp.hstack([jnp.ones((N,1)), income])

    def simulate_binary_logit(x, beta):
      beta = beta.reshape(-1,1)
      N = x.shape[0]
      J = beta.shape[0]

      epsilon = jax.random.gumbel(rng,shape =(N,J))
      Beta_augmented = jnp.hstack([beta, jnp.zeros_like(beta)])
      utility = x @ Beta_augmented + epsilon

      choice_idx = onp.argmax(utility, axis=1)
      return (choice_idx).reshape(-1,1)

    y = simulate_binary_logit(x, beta)
    y = jnp.ravel(y)

    # numpy version
    def binary_logit_log_likelihood(beta, y,x):
      lambda_xb = onp.exp(x@beta) / (1 + onp.exp(x@beta))
      ll_i = y * onp.log(lambda_xb) + (1-y) * onp.log(1-lambda_xb)
      ll = -onp.sum(ll_i)
      return ll

    # jax version
    def binary_logit_log_likelihood_jax(beta, y, x):
      lambda_xb = jnp.exp(x@beta) / (1 + jnp.exp(x@beta))
      ll_i = y * jnp.log(lambda_xb) + (1-y) * jnp.log(1-lambda_xb)
      ll = -jnp.sum(ll_i)
      return ll

    beta_init = jnp.array([0.01,0.01])

    # using scipy
    scipy_res = scipy_opt.minimize(
      fun=binary_logit_log_likelihood, 
      args=(onp.asarray(y),onp.asarray(x)), 
      x0 = (onp.asarray(beta_init)), method='BFGS'
    ).x

    # using jaxopt
    solver = LBFGS(fun=binary_logit_log_likelihood_jax, maxiter=100,
                   linesearch="zoom", maxls=10, tol=1e-12)
    jaxopt_res = solver.run(beta_init, y, x).params

    # comparison
    scipy_val = binary_logit_log_likelihood(scipy_res,
                                            onp.asarray(y),
                                            onp.asarray(x))
    jaxopt_val = binary_logit_log_likelihood(jaxopt_res, y, x)
    self.assertLessEqual(jaxopt_val, scipy_val + 3e-5)

  @parameterized.product(linesearch=['zoom', 'backtracking', 'hager-zhang'])
  def test_complex(self, linesearch):
    """Test that optimization over complex variable z = x + jy matches equivalent real case"""
    # NOTE(vroulet): At high precision, the results differ slightly 
    # (tol=5*1e-15 instead of tol=1e-15 by default at double precision)
    tol = 5*1e-15 if jax.config.jax_enable_x64 else None

    W = jnp.array(
      [[1, - 2],
       [3, 4],
       [-4 + 2j, 5 - 3j],
       [-2 - 2j, 6]]
    )

    def C2R(z):
      return jnp.stack((z.real, z.imag))

    def R2C(x):
      return x[..., 0, :] + 1j * x[..., 1, :]  # if len(x)>0 else jnp.zeros_like(x)

    def f(z):
      return W @ z

    def loss_complex(z):
      return jnp.sum(jnp.abs(f(z)) ** 1.5)

    def loss_real(zR):
      return loss_complex(R2C(zR))

    z0 = jnp.array([1 - 1j, 0 + 1j])

    solver_options = dict(maxiter=5,
                          history_size=3,
                          stepsize=-1,
                          linesearch=linesearch,
                          min_stepsize=0.1,
                          max_stepsize=2,
                          )
    solver_C = LBFGS(fun=loss_complex, **solver_options)
    solver_R = LBFGS(fun=loss_real, **solver_options)
    sol_C, state_C = solver_C.run(z0)
    sol_R, state_R = solver_R.run(C2R(z0))

    self.assertArraysAllClose(sol_C, R2C(sol_R), atol=tol, rtol=tol)
    self.assertArraysAllClose(state_C.s_history, R2C(state_R.s_history), atol=tol, rtol=tol)

  def test_handling_pytrees(self):
    def fun_(x):
      return sum(100.0 * (x[..., 1:] - x[..., :-1] ** 2.0) ** 2.0 + (1 - x[..., :-1]) ** 2.0)

    def fun(x):
      return tree_sum(tree_map(fun_, x))

    key = random.PRNGKey(0)
    x_arr0 = random.normal(key, (2, 4))
    x_tree0 = (x_arr0[0], x_arr0[1])

    lbfgs = LBFGS(fun=fun, maxiter=5)
    x_arr, _ = lbfgs.run(x_arr0)
    x_tree, _ = lbfgs.run(x_tree0)
    x_tree = jnp.stack((x_tree[0], x_tree[1]))
    self.assertArraysAllClose(x_arr, x_tree)

  def test_zero_history(self):
    # Should match a gradient descent if history_size = 0
    def fun(x):
      return sum(100.0*(x[1:] - x[:-1]**2.0)**2.0 + (1 - x[:-1])**2.0)
    
    x0 = jnp.zeros(2)
    lbfgs = LBFGS(fun=fun, tol=1e-3, maxiter=20, history_size=0, stepsize=0.01)
    x_lbfgs = lbfgs.run(x0).params
    gd = GradientDescent(fun=fun, tol=1e-3, maxiter=20, stepsize=0.01, acceleration=False)
    x_gd = gd.run(x0).params
    self.assertArraysAllClose(x_lbfgs, x_gd)

  @parameterized.product(
    fun_init_and_opt=[
      ('rosenbrock', onp.zeros(2), 0.),
      ('himmelblau', onp.ones(2), 0.),
      ('matyas', onp.ones(2) * 6., 0.),
      ('eggholder', onp.ones(2) * 100., None),  
      ('zakharov', onp.array([600.0, 700.0, 200.0, 100.0, 90.0, 1e4]), 0.),
    ],
  )
  def test_against_scipy(self, fun_init_and_opt):
    # Taken from previous jax tests
    # NOTE(vroulet): Unclear whether lbfgs or bfgs can find true minimum for
    # eggholder, but they seem to converge to the same solution with current 
    # implementation and initialization, which is a good check.
    tol = 1e-15 if jax.config.jax_enable_x64 else 1e-7
    fun_name, x0, opt = fun_init_and_opt
    jnp_fun, onp_fun = get_fun(fun_name, jnp), get_fun(fun_name, onp)
    jaxopt_res = LBFGS(jnp_fun, tol=tol).run(x0).params
    scipy_res = scipy_opt.minimize(onp_fun, x0, method='BFGS').x
    # scipy not good for matyas and zakharov functions, 
    # compare to true minimum, zero
    if fun_name in ['matyas', 'zakharov']:
      self.assertAllClose(jaxopt_res, jnp.zeros_like(jaxopt_res))
    elif fun_name == 'eggholder' and jax.config.jax_enable_x64:
      # FIXME(vroulet): at high precision, eggholder poses issues,
      # even in function values
      self.assertAllClose(jnp_fun(jaxopt_res), onp_fun(scipy_res), rtol=1e-14)
    else:
      self.assertAllClose(jaxopt_res, scipy_res, check_dtypes=False)
    if opt is not None:
      self.assertAllClose(jnp_fun(jaxopt_res), opt, check_dtypes=False)

  def test_minimize_bad_initial_values(self):
    # Taken from previous jax tests
    # This test runs deliberately "bad" initial values to test that handling
    # of failed line search, etc. is the same across implementations
    tol = 1e-6 if jax.config.jax_enable_x64 else 1e-3
    initial_value = onp.array([92, 0.001])
    jnp_fun, onp_fun = get_fun('himmelblau', jnp), get_fun('himmelblau', onp)
    # Here reuqires higher number of linesearch for jaxopt
    jaxopt_res = LBFGS(jnp_fun, tol=tol).run(initial_value).params
    scipy_res = scipy_opt.minimize(
        fun=onp_fun,
        jac=jax.grad(onp_fun),
        method='BFGS',
        x0=initial_value,
    ).x
    # Scipy and jaxopt converge to different minima so check in function values
    self.assertAllClose(onp_fun(scipy_res), jnp_fun(jaxopt_res))
    self.assertAllClose(jnp_fun(jaxopt_res), 0.)

  def test_steep(self):
    # Taken from previous jax tests
    # See jax related issue https://github.com/google/jax/issues/4594
    n = 2
    A = jnp.eye(n) * 1e4
    def fun(x):
      return jnp.mean((A @ x) ** 2)
    results = LBFGS(fun).run(init_params=jnp.ones(n)).params
    self.assertAllClose(results, jnp.zeros(n))

  def test_minimize_complex_sphere(self):
    # Taken from previous jax tests
    z0 = jnp.array([1., 2. - 3.j, 4., -5.j])

    def fun(z):
      return jnp.real(jnp.dot(jnp.conj(z - z0), z - z0))

    jaxopt_res = LBFGS(fun, tol=1e-6).run(jnp.zeros_like(z0)).params

    self.assertAllClose(jaxopt_res, z0)

  def test_complex_rosenbrock(self):
    # Taken from previous jax tests
    tol = 1e-12 if jax.config.jax_enable_x64 else 1e-6

    complex_dim = 5

    fun_real = get_fun('rosenbrock', jnp)
    init_re = jnp.zeros((2 * complex_dim,), dtype=complex)
    expect_re = jnp.ones((2 * complex_dim,), dtype=complex)

    def fun(z):
      x_re = jnp.concatenate([jnp.real(z), jnp.imag(z)])
      return fun_real(x_re)

    init = init_re[:complex_dim] + 1.j * init_re[complex_dim:]
    expect = expect_re[:complex_dim] + 1.j * expect_re[complex_dim:]

    jaxopt_res = LBFGS(fun, tol=tol).run(init).params
    self.assertAllClose(jaxopt_res, expect)

if __name__ == '__main__':
  absltest.main()



"""Utility functions for applying identifiability constraints to rank deficient feature matrices."""

from __future__ import annotations

from functools import partial
from typing import Callable, Tuple

import jax
import jax.numpy as jnp
import numpy as np
from jax.typing import ArrayLike as JaxArray
from numpy.typing import NDArray

from .basis._basis import Basis
from .tree_utils import get_valid_multitree, tree_slice
from .type_casting import support_pynapple
from .validation import _warn_if_not_float64

_WARN_FLOAT32_MESSAGE = (
    "The feature matrix is not of dtype `float64`. Consider converting it to `float64` "
    "for increased numerical precision when computing the matrix rank. You can enable "
    "float64 precision globally by adding:\n\n    jax.config.update('jax_enable_x64', True)\n"
)


def add_constant(x):
    """Add intercept term."""
    return jnp.hstack((jnp.ones((x.shape[0], 1), dtype=x.dtype), x))


@partial(jax.jit, static_argnums=(2,))
def _drop_and_compute_rank(feature_matrix, idx, preprocessing_func=add_constant):
    """Drop column idx and compute rank."""
    feature_dropped = feature_matrix.at[:, idx].set(0.0)
    rank_after_drop_column = jnp.linalg.matrix_rank(preprocessing_func(feature_dropped))
    return feature_dropped, rank_after_drop_column


@partial(jax.jit, static_argnums=(1, 2, 3))
def _find_drop_column(
    feature_matrix: JaxArray,
    rank: int,
    max_drop: int,
    preprocessing_func: Callable = add_constant,
) -> JaxArray:
    """
    Find a minimal subset linearly dependent columns that can be dropped.

    This function loops over the columns of a matrix and checks if each column is linearly dependent from the others.
    If the i-th column is linearly dependent, then drop_cols[i] is set to True, and feature_matrix[:, i] is set to 0.
    The loop is stopped when max_drop linearly dependent columns are found.

    Parameters
    ----------
    feature_matrix:
        The rank deficient feature matrix.
    rank:
        The rank of the matrix.
    max_drop:
        Number of columns to be dropped.
    preprocessing_func:
        Additional processing of the feature matrix. By default, add an intercept term. Other processing could
        entail mean-centering the columns or similar.

    Returns
    -------
    drop_cols:
        A boolean vector, True if the column should be dropped, False otherwise.

    """

    def drop_col_and_update(features, dropped_features, drop_cols, iter_num):
        """Drop feature and update drop column boolean."""
        return dropped_features, drop_cols.at[iter_num].set(True)

    def do_not_drop(features, dropped_features, drop_cols, iter_num):
        """Do not drop."""
        return features, drop_cols

    def check_column(iter_num, state):
        """Drop a column if rank is not affected."""
        matrix, _, original_rank, drop_cols, mx_drop = state
        # drop the column (by set to zero) and compute rank
        col_dropped_matrix, mat_rank = _drop_and_compute_rank(
            matrix, iter_num, preprocessing_func
        )
        # apply the change if rank stays constant, do nothing otherwise.
        matrix, drop_cols = jax.lax.cond(
            mat_rank == original_rank,  # condition
            drop_col_and_update,  # true function
            do_not_drop,  # false function
            matrix,  # parameters
            col_dropped_matrix,
            drop_cols,
            iter_num,
        )
        return matrix, mat_rank, original_rank, drop_cols, mx_drop

    def body_func(iter_num, state):
        drop_cols, max_drop = state[-2:]
        return jax.lax.cond(
            drop_cols.sum() < max_drop, check_column, lambda it, x: x, iter_num, state
        )

    init_state = (
        feature_matrix,
        jnp.array(0),
        jnp.array(rank),
        jnp.zeros(feature_matrix.shape[1], dtype=bool),
        max_drop,
    )
    final_state = jax.lax.fori_loop(0, feature_matrix.shape[1], body_func, init_state)

    return final_state[3]


def _add_invalid_entries(feature_matrix, shape_first_axis, is_valid):
    """Add invalid entries to match original shape."""
    feature_matrix = (
        jnp.full(
            (shape_first_axis, *feature_matrix.shape[1:]),
            jnp.nan,
            dtype=feature_matrix.dtype,
        )
        .at[is_valid]
        .set(feature_matrix)
    )
    return feature_matrix


def _apply_identifiability_constraints(
    feature_matrix: JaxArray,
    preprocessing_func: Callable = add_constant,
    warn_if_float32: bool = True,
) -> Tuple[JaxArray, JaxArray]:
    """
    Apply identifiability constraints to a design matrix `feature_matrix`.

    Private function that does the actual computation on a single feature_matrix.
    """
    if warn_if_float32:
        _warn_if_not_float64(feature_matrix, _WARN_FLOAT32_MESSAGE)

    shape_sample_axis = feature_matrix.shape[0]
    is_valid = get_valid_multitree(feature_matrix)

    # compute initial rank if needed
    feature_matrix = tree_slice(feature_matrix, is_valid)
    feature_matrix_with_intercept = preprocessing_func(feature_matrix)
    rank = jnp.linalg.matrix_rank(feature_matrix_with_intercept)

    # full rank, no extra computation needed
    if rank == feature_matrix_with_intercept.shape[1]:
        feature_matrix = _add_invalid_entries(
            feature_matrix, shape_sample_axis, is_valid
        )
        return feature_matrix, jnp.zeros((feature_matrix.shape[1]), dtype=bool)

    max_drop = feature_matrix_with_intercept.shape[1] - rank

    # run the search
    drop_cols = _find_drop_column(
        feature_matrix,
        rank=int(rank),
        max_drop=int(max_drop),
        preprocessing_func=preprocessing_func,
    )

    # return the output matrix and the dropped indices
    feature_matrix = _add_invalid_entries(
        feature_matrix[:, ~drop_cols],
        shape_sample_axis,
        is_valid,
    )
    return feature_matrix, drop_cols


@support_pynapple(conv_type="jax")
def apply_identifiability_constraints(
    feature_matrix: NDArray | JaxArray,
    add_intercept: bool = True,
    warn_if_float32: bool = True,
) -> Tuple[NDArray, NDArray[int]]:
    """
    Apply identifiability constraints to a design matrix ``X``.

    Removes columns from ``X`` until it is full rank to ensure the uniqueness
    of the GLM (Generalized Linear Model) maximum-likelihood solution. This is particularly
    crucial for models using bases like BSplines and CyclicBspline, which, due to their
    construction, sum to 1 and can cause rank deficiency when combined with an intercept.

    For GLMs, this rank deficiency means that different sets of coefficients might yield
    identical predicted rates and log-likelihoods, complicating parameter learning, especially
    in the absence of regularization.

    For very large feature matrices generated by a sum of low-dimensional basis components, consider
    ``apply_identifiability_constraints_by_basis_component``.

    Parameters
    ----------
    feature_matrix:
        The design matrix before applying the identifiability constraints.
    add_intercept:
        Set to True if your model will add an intercept term, False otherwise.
    warn_if_float32:
        Raise a warning if feature matrix dtype is float32.

    Returns
    -------
    constrained_x:
        The adjusted design matrix with redundant columns dropped and columns mean-centered.
    kept_columns:
        The columns that have been kept.

    Examples
    --------
    >>> import numpy as np
    >>> from nemos.identifiability_constraints import apply_identifiability_constraints
    >>> from nemos.basis import BSplineEval
    >>> from nemos.glm import GLM
    >>> import jax
    >>> jax.config.update('jax_enable_x64', True)
    >>> # define a feature matrix
    >>> bas = BSplineEval(5) + BSplineEval(6)
    >>> feature_matrix = bas.compute_features(np.random.randn(100), np.random.randn(100))
    >>> # apply constraints
    >>> constrained_x, kept_columns = apply_identifiability_constraints(feature_matrix)
    >>> constrained_x.shape
    (100, 9)
    >>> kept_columns
    array([ 1,  2,  3,  4,  6,  7,  8,  9, 10])

    Notes
    -----
    Compilation is triggered at every loop. This can be slower than pure python for low number
    of samples and low dimension for the feature matrix.
    Usually, the design matrices we work with have a large number of samples.
    Running the code on GPU will reduce the computation time significantly.
    """
    if add_intercept:
        preproc_design = add_constant
    else:

        def preproc_design(x):
            return x

    # return the output matrix and the dropped indices
    constrained_x, discarded_columns = _apply_identifiability_constraints(
        jnp.asarray(feature_matrix),
        preprocessing_func=preproc_design,
        warn_if_float32=warn_if_float32,
    )
    kept_columns = np.arange(feature_matrix.shape[1])[~discarded_columns]
    return np.asarray(constrained_x), kept_columns


@support_pynapple(conv_type="jax")
def apply_identifiability_constraints_by_basis_component(
    basis: Basis,
    feature_matrix: NDArray,
    add_intercept: bool = True,
) -> Tuple[NDArray, NDArray]:
    """Apply identifiability constraint to a design matrix for each component of an additive basis.

    Parameters
    ----------
    basis:
        The basis that computed ``feature_matrix``.
    feature_matrix:
        The feature matrix before applying the identifiability constraints.
    add_intercept:
        Set to True if your model will add an intercept term, False otherwise.

    Returns
    -------
    constrained_x:
        The adjusted feature matrix after applying the identifiability constraints as numpy array.
    kept_columns:
        Indices of the columns that are kept. This should be used for applying the same transformation
        to a feature matrix generated from different a set of inputs (as for a test set).

    Examples
    --------
    >>> import numpy as np
    >>> import jax
    >>> from nemos.identifiability_constraints import apply_identifiability_constraints_by_basis_component
    >>> from nemos.basis import BSplineEval
    >>> from nemos.glm import GLM
    >>> jax.config.update('jax_enable_x64', True)
    >>> # define a feature matrix
    >>> bas = BSplineEval(5) + BSplineEval(6)
    >>> feature_matrix = bas.compute_features(np.random.randn(100), np.random.randn(100))
    >>> # apply constraints
    >>> constrained_x, kept_columns = apply_identifiability_constraints_by_basis_component(bas, feature_matrix)
    >>> constrained_x.shape
    (100, 9)
    >>> # generate a test set, shape (20, 11)
    >>> test_x = bas.compute_features(np.random.randn(20), np.random.randn(20))
    >>> test_x.shape
    (20, 11)
    >>> # apply constraint to test set
    >>> test_x = test_x[:, kept_columns]
    >>> test_x.shape
    (20, 9)
    >>> # fit on train and predict on test set
    >>> rate = GLM().fit(constrained_x, np.random.poisson(size=100)).predict(test_x)

    """
    # gets a dictionary with feature specific feature matrices
    # stored in tensors of shape (n_samples, n_inputs, n_features)
    # n_inputs can be larger than one if basis is used to perform
    # convolutions on multiple signals (as for counts TsdFrames)
    splits_x = basis.split_by_feature(feature_matrix)

    # list the arrays
    split_x = jax.tree_util.tree_leaves(splits_x)
    # add dim if needed (the dim is at least 2, (n_samples, n_basis)
    split_x = [x if x.ndim > 2 else x[:, None] for x in split_x]
    # flatten over inputs
    split_x = [x.reshape(x.shape[0], -1, x.shape[-1]) for x in split_x]
    # list leaves and unwrap over input dimension. Additive components have shapes:
    # (n_samples, n_inputs, n_basis_funcs)
    split_by_input_x = [x[:, k] for x in split_x for k in range(x.shape[1])]

    apply_identifiability = partial(
        apply_identifiability_constraints,
        add_intercept=add_intercept,
        warn_if_float32=False,
    )

    _warn_if_not_float64(split_by_input_x, _WARN_FLOAT32_MESSAGE)

    constrained_x_and_columns = jax.tree_util.tree_map(
        apply_identifiability, split_by_input_x
    )

    # unpack the outputs into array and dropped colum indices
    def is_leaf(x):
        return isinstance(x, tuple)

    constrained_x = tree_slice(constrained_x_and_columns, idx=0, is_leaf=is_leaf)
    kept_columns = tree_slice(constrained_x_and_columns, idx=1, is_leaf=is_leaf)

    # stack the arrays back into a feature matrix
    constrained_x = np.hstack(constrained_x)

    # indices are referenced to the sub-matrices, get the absolute index in the feature matrix
    # calculate the shifts for each component
    shifts = list(
        np.cumsum(
            [0]
            + [
                sub_x.shape[1]
                for sub_x in jax.tree_util.tree_leaves(split_by_input_x)[:-1]
            ]
        )
    )
    kept_columns = jax.tree_util.tree_map(lambda x, y: x + y, kept_columns, shifts)
    kept_columns = np.hstack(kept_columns)
    return constrained_x, kept_columns



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools

from absl.testing import absltest

import jax
import jax.numpy as jnp

from nemos.third_party.jaxopt.jaxopt import implicit_diff as idf
from nemos.third_party.jaxopt.jaxopt._src import test_util

from sklearn import datasets


def ridge_objective(params, lam, X, y):
  residuals = jnp.dot(X, params) - y
  return 0.5 * jnp.mean(residuals ** 2) + 0.5 * lam * jnp.sum(params ** 2)


# def ridge_solver(init_params, lam, X, y):
def ridge_solver(init_params, lam, X, y):
  del init_params  # not used
  XX = jnp.dot(X.T, X)
  Xy = jnp.dot(X.T, y)
  I = jnp.eye(X.shape[1])
  return jnp.linalg.solve(XX + lam * len(y) * I, Xy)


def ridge_solver_with_kwargs(init_params, **kw):
  return ridge_solver(init_params, kw['lam'], kw['X'], kw['y'])


class ImplicitDiffTest(test_util.JaxoptTestCase):

  def test_root_vjp(self):
    X, y = datasets.make_regression(n_samples=10, n_features=3, random_state=0)
    optimality_fun = jax.grad(ridge_objective)
    lam = 5.0
    sol = ridge_solver(None, lam, X, y)
    vjp = lambda g: idf.root_vjp(optimality_fun=optimality_fun,
                                 sol=sol,
                                 args=(lam, X, y),
                                 cotangent=g)[0]  # vjp w.r.t. lam
    I = jnp.eye(len(sol))
    J = jax.vmap(vjp)(I)
    J_num = test_util.ridge_solver_jac(X, y, lam, eps=1e-4)
    self.assertArraysAllClose(J, J_num, atol=5e-2)

  def test_root_jvp(self):
    X, y = datasets.make_regression(n_samples=10, n_features=3, random_state=0)
    optimality_fun = jax.grad(ridge_objective)
    lam = 5.0
    sol = ridge_solver(None, lam, X, y)
    J = idf.root_jvp(optimality_fun=optimality_fun,
                     sol=sol,
                     args=(lam, X, y),
                     tangents=(1.0, jnp.zeros_like(X), jnp.zeros_like(y)))
    J_num = test_util.ridge_solver_jac(X, y, lam, eps=1e-4)
    self.assertArraysAllClose(J, J_num, atol=5e-2)

  def test_custom_root(self):
    X, y = datasets.make_regression(n_samples=10, n_features=3, random_state=0)
    optimality_fun = jax.grad(ridge_objective)
    lam = 5.0
    ridge_solver_decorated = idf.custom_root(optimality_fun)(ridge_solver)
    sol = ridge_solver(None, lam=lam, X=X, y=y)
    sol_decorated = ridge_solver_decorated(None, lam=lam, X=X, y=y)
    self.assertArraysAllClose(sol, sol_decorated, atol=1e-4)
    J_num = test_util.ridge_solver_jac(X, y, lam, eps=1e-4)
    J = jax.jacrev(ridge_solver_decorated, argnums=1)(None, lam, X=X, y=y)
    self.assertArraysAllClose(J, J_num, atol=5e-2)

  def test_custom_root_with_has_aux(self):
    def ridge_solver_with_aux(init_params, lam, X, y):
      return ridge_solver(init_params, lam, X, y), None

    X, y = datasets.make_regression(n_samples=10, n_features=3, random_state=0)
    optimality_fun = jax.grad(ridge_objective)
    lam = 5.0
    decorator = idf.custom_root(optimality_fun, has_aux=True)
    ridge_solver_decorated = decorator(ridge_solver_with_aux)
    sol = ridge_solver(None, lam=lam, X=X, y=y)
    sol_decorated = ridge_solver_decorated(None, lam=lam, X=X, y=y)[0]
    self.assertArraysAllClose(sol, sol_decorated, atol=1e-4)
    J_num = test_util.ridge_solver_jac(X, y, lam, eps=1e-4)
    J, _ = jax.jacrev(ridge_solver_decorated, argnums=1)(None, lam, X=X, y=y)
    self.assertArraysAllClose(J, J_num, atol=5e-2)

  def test_custom_fixed_point(self):
    X, y = datasets.make_regression(n_samples=10, n_features=3, random_state=0)
    grad_fun = jax.grad(ridge_objective)
    fp_fun = lambda x, *args: x - grad_fun(x, *args)
    fp_fun = functools.wraps(grad_fun)(fp_fun)  # carries grad_fun signature
    lam = 5.0
    ridge_solver_decorated = idf.custom_fixed_point(fp_fun)(ridge_solver)
    sol = ridge_solver(None, lam=lam, X=X, y=y)
    sol_decorated = ridge_solver_decorated(None, lam=lam, X=X, y=y)
    self.assertArraysAllClose(sol, sol_decorated, atol=1e-4)
    J_num = test_util.ridge_solver_jac(X, y, lam, eps=1e-4)
    J = jax.jacrev(ridge_solver_decorated, argnums=1)(None, lam, X=X, y=y)
    self.assertArraysAllClose(J, J_num, atol=5e-2)

  def test_custom_root_with_kwargs(self):
    X, y = datasets.make_regression(n_samples=10, n_features=3, random_state=0)
    lam = 5.0
    w_ref = ridge_solver(None, lam, X, y)

    cr_solver = idf.custom_root(jax.grad(ridge_objective))(ridge_solver_with_kwargs)
    w = cr_solver(None, lam=lam, X=X, y=y)
    w_jit = jax.jit(cr_solver)(None, lam=lam, X=X, y=y)

    self.assertArraysAllClose(w_ref, w, atol=5e-2)
    self.assertArraysAllClose(w_jit, w, atol=5e-2)

  def test_custom_root_with_kwargs_unordered(self):
    X, y = datasets.make_regression(n_samples=10, n_features=3, random_state=0)
    lam = 5.0
    w_ref = ridge_solver(None, lam, X, y)

    cr_solver = idf.custom_root(jax.grad(ridge_objective))(ridge_solver_with_kwargs)
    w = cr_solver(None, X=X, lam=lam, y=y)
    w_jit = jax.jit(cr_solver)(None, X=X, lam=lam, y=y)

    self.assertArraysAllClose(w_ref, w, atol=5e-2)
    self.assertArraysAllClose(w_jit, w, atol=5e-2)

  def test_custom_root_with_kwargs_grad(self):
    X, y = datasets.make_regression(n_samples=10, n_features=3, random_state=0)
    lam = 5.0
    lam_grad_ref = jax.grad(lambda lam: ridge_solver(None, lam, X, y)[0])(lam)

    cr_solver = idf.custom_root(jax.grad(ridge_objective))(ridge_solver_with_kwargs)
    grad_fun = jax.grad(lambda lam: cr_solver(None, lam=lam, X=X, y=y)[0])
    lam_grad = grad_fun(lam)
    lam_grad_jit = jax.jit(grad_fun)(lam)

    self.assertArraysAllClose(lam_grad_ref, lam_grad, atol=5e-2)
    self.assertArraysAllClose(lam_grad_jit, lam_grad, atol=5e-2)

  def test_custom_root_with_kwargs_grad_unordered(self):
    X, y = datasets.make_regression(n_samples=10, n_features=3, random_state=0)
    lam = 5.0
    lam_grad_ref = jax.grad(lambda lam: ridge_solver(None, lam, X, y)[0])(lam)

    cr_solver = idf.custom_root(jax.grad(ridge_objective))(ridge_solver_with_kwargs)
    grad_fun = jax.grad(lambda lam: cr_solver(None, X=X, lam=lam, y=y)[0])
    lam_grad = grad_fun(lam)
    lam_grad_jit = jax.jit(grad_fun)(lam)

    self.assertArraysAllClose(lam_grad_ref, lam_grad, atol=5e-2)
    self.assertArraysAllClose(lam_grad_jit, lam_grad, atol=5e-2)

if __name__ == '__main__':
  # Uncomment the line below in order to run in float64.
  # jax.config.update("jax_enable_x64", True)
  absltest.main()



"""Base class defining the interface for solvers that can be used by `BaseRegressor`."""

import abc
from typing import Any, Callable, Generic, NamedTuple

from ..regularizer import Regularizer
from ..typing import Params, SolverState, StepResult


class OptimizationInfo(NamedTuple):
    """Basic diagnostic information about finished optimization runs."""

    # Not all JAXopt solvers store the function value.
    # None means missing value, while NaN usually indicates a diverged optimization
    function_val: float | None
    num_steps: int
    converged: bool
    reached_max_steps: bool


class AbstractSolver(abc.ABC, Generic[SolverState]):
    """
    Base class defining the interface for solvers that can be used by `BaseRegressor`.

    All solver parameters (e.g. tolerance, number of steps) are passed to `__init__`,
    the other methods only take parameters, solver state, and the positional arguments of
    the objective function.
    """

    @abc.abstractmethod
    def __init__(
        self,
        unregularized_loss: Callable,
        regularizer: Regularizer,
        regularizer_strength: float | None,
        **solver_init_kwargs,
    ):
        """
        Create the solver.

        Arguments
        ---------
        unregularized_loss:
            Unregularized loss function.
            Currently `BaseRegressor._predict_and_compute_loss`.
        regularizer:
            Regularizer object used to create the penalized loss
            or get the proximal operator from.
        regularizer_strength:
            Regularizer strength.
        **solver_init_kwargs:
            Keyword arguments modifying the solver's behavior.
        """
        pass

    @abc.abstractmethod
    def init_state(self, init_params: Params, *args: Any) -> SolverState:
        """
        Initialize the solver state.

        Used by `BaseRegressor.initialize_state`
        """
        pass

    @abc.abstractmethod
    def update(self, params: Params, state: SolverState, *args: Any) -> StepResult:
        """
        Perform a single step/update of the optimization process.

        Used by `BaseRegressor.update`.
        """
        pass

    @abc.abstractmethod
    def run(self, init_params: Params, *args: Any) -> StepResult:
        """
        Run a full optimization process until a stopping criterion is reached.

        Used by `BaseRegressor.fit`.
        """
        pass

    @classmethod
    @abc.abstractmethod
    def get_accepted_arguments(cls) -> set[str]:
        """
        Set of argument names accepted by the solver.

        Used by `BaseRegressor` to determine what arguments
        can be passed to the solver's __init__.
        """
        pass

    @abc.abstractmethod
    def get_optim_info(self, state: SolverState) -> OptimizationInfo:
        """Extract some commong info about the optimization process.

        Currently, the following info is extracted:
        - final function value (where available)
        - number of steps
        - whether the optimization converged
        - whether the max number of steps were reached
        """
        pass



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Linear Operator tests."""

from absl.testing import absltest

import jax.numpy as jnp
import numpy as onp

from nemos.third_party.jaxopt.jaxopt._src.linear_operator import FunctionalLinearOperator
from nemos.third_party.jaxopt.jaxopt._src import test_util


class LinearOperatorTest(test_util.JaxoptTestCase):

  def test_matvec_and_rmatvec(self):
    rng = onp.random.RandomState(0)
    A = rng.randn(5, 4)
    matvec = lambda A,x: jnp.dot(A, x)
    x = rng.randn(4)
    y = rng.randn(5)
    linop_A = FunctionalLinearOperator(matvec, A)
    mv_A, rmv_A = linop_A.matvec_and_rmatvec(x, y)
    self.assertArraysAllClose(mv_A, jnp.dot(A, x))
    self.assertArraysAllClose(rmv_A, jnp.dot(A.T, y))


if __name__ == '__main__':
  absltest.main()



import jax.numpy as jnp
import pytest

from nemos.proximal_operator import _vmap_norm2_masked_2, prox_group_lasso, prox_lasso


@pytest.mark.parametrize("prox_operator", [prox_group_lasso, prox_lasso])
def test_prox_operator_returns_tuple(prox_operator, example_data_prox_operator):
    """Test whether the proximal operator returns a tuple."""
    args = example_data_prox_operator
    args = args if prox_operator is prox_group_lasso else (*args[:2], *args[3:])
    params_new = prox_operator(*args)
    assert isinstance(params_new, tuple)


@pytest.mark.parametrize("prox_operator", [prox_group_lasso, prox_lasso])
def test_prox_operator_returns_tuple_multineuron(
    prox_operator, example_data_prox_operator_multineuron
):
    """Test whether the tuple returned by the proximal operator has a length of 2."""
    args = example_data_prox_operator_multineuron
    args = args if prox_operator is prox_group_lasso else (*args[:2], *args[3:])
    params_new = prox_operator(*args)
    assert isinstance(params_new, tuple)


@pytest.mark.parametrize("prox_operator", [prox_group_lasso, prox_lasso])
def test_prox_operator_tuple_length(prox_operator, example_data_prox_operator):
    """Test whether the tuple returned by the proximal operator has a length of 2."""
    args = example_data_prox_operator
    args = args if prox_operator is prox_group_lasso else (*args[:2], *args[3:])
    params_new = prox_operator(*args)
    assert len(params_new) == 2


@pytest.mark.parametrize("prox_operator", [prox_group_lasso, prox_lasso])
def test_prox_operator_tuple_length_multineuron(
    prox_operator, example_data_prox_operator_multineuron
):
    """Test whether the tuple returned by the proximal operator has a length of 2."""
    args = example_data_prox_operator_multineuron
    args = args if prox_operator is prox_group_lasso else (*args[:2], *args[3:])
    params_new = prox_operator(*args)
    assert len(params_new) == 2


@pytest.mark.parametrize("prox_operator", [prox_group_lasso, prox_lasso])
def test_prox_operator_weights_shape(prox_operator, example_data_prox_operator):
    """Test whether the shape of the weights in the proximal operator is correct."""
    args = example_data_prox_operator
    args = args if prox_operator is prox_group_lasso else (*args[:2], *args[3:])
    params_new = prox_operator(*args)
    assert params_new[0].shape == args[0][0].shape


@pytest.mark.parametrize("prox_operator", [prox_group_lasso, prox_lasso])
def test_prox_operator_weights_shape_multineuron(
    prox_operator, example_data_prox_operator_multineuron
):
    """Test whether the shape of the weights in the proximal operator is correct."""
    args = example_data_prox_operator_multineuron
    args = args if prox_operator is prox_group_lasso else (*args[:2], *args[3:])
    params_new = prox_operator(*args)
    assert params_new[0].shape == args[0][0].shape


@pytest.mark.parametrize("prox_operator", [prox_group_lasso, prox_lasso])
def test_prox_operator_intercepts_shape(prox_operator, example_data_prox_operator):
    """Test whether the shape of the intercepts in the proximal operator is correct."""
    args = example_data_prox_operator
    args = args if prox_operator is prox_group_lasso else (*args[:2], *args[3:])
    params_new = prox_operator(*args)
    assert params_new[1].shape == args[0][1].shape


@pytest.mark.parametrize("prox_operator", [prox_group_lasso, prox_lasso])
def test_prox_operator_intercepts_shape_multineuron(
    prox_operator, example_data_prox_operator_multineuron
):
    """Test whether the shape of the intercepts in the proximal operator is correct."""
    args = example_data_prox_operator_multineuron
    args = args if prox_operator is prox_group_lasso else (*args[:2], *args[3:])
    params_new = prox_operator(*args)
    assert params_new[1].shape == args[0][1].shape


def test_vmap_norm2_masked_2_returns_array(example_data_prox_operator):
    """Test whether _vmap_norm2_masked_2 returns a NumPy array."""
    params, _, mask, _ = example_data_prox_operator
    l2_norm = _vmap_norm2_masked_2(params[0], mask)
    assert isinstance(l2_norm, jnp.ndarray)


def test_vmap_norm2_masked_2_shape(example_data_prox_operator):
    """Test whether the shape of the result from _vmap_norm2_masked_2 is correct."""
    params, _, mask, _ = example_data_prox_operator
    l2_norm = _vmap_norm2_masked_2(params[0], mask)
    assert l2_norm.shape == (params[0].shape[0], mask.shape[0])


def test_vmap_norm2_masked_2_shape_multineuron(example_data_prox_operator_multineuron):
    """Test whether the shape of the result from _vmap_norm2_masked_2 is correct."""
    params, _, mask, _ = example_data_prox_operator_multineuron
    l2_norm = _vmap_norm2_masked_2(params[0].T, mask)
    assert l2_norm.shape == (params[0].shape[1], mask.shape[0])


def test_vmap_norm2_masked_2_non_negative(example_data_prox_operator):
    """Test whether all elements of the result from _vmap_norm2_masked_2 are non-negative."""
    params, _, mask, _ = example_data_prox_operator
    l2_norm = _vmap_norm2_masked_2(params[0], mask)
    assert jnp.all(l2_norm >= 0)


def test_vmap_norm2_masked_2_non_negative_multineuron(
    example_data_prox_operator_multineuron,
):
    """Test whether all elements of the result from _vmap_norm2_masked_2 are non-negative."""
    params, _, mask, _ = example_data_prox_operator_multineuron
    l2_norm = _vmap_norm2_masked_2(params[0].T, mask)
    assert jnp.all(l2_norm >= 0)


def test_prox_operator_shrinks_only_masked(example_data_prox_operator):
    params, _, mask, _ = example_data_prox_operator
    mask = mask.at[:, 1].set(jnp.zeros(2))
    params_new = prox_group_lasso(params, 0.05, mask)
    assert params_new[0][1] == params[0][1]
    assert all(params_new[0][i] < params[0][i] for i in [0, 2, 3])


def test_prox_operator_shrinks_only_masked_multineuron(
    example_data_prox_operator_multineuron,
):
    params, _, mask, _ = example_data_prox_operator_multineuron
    mask = mask.astype(float)
    mask = mask.at[:, 1].set(jnp.zeros(2))
    params_new = prox_group_lasso(params, 0.05, mask)
    assert jnp.all(params_new[0][1] == params[0][1])
    assert all(jnp.all(params_new[0][i] < params[0][i]) for i in [0, 2, 3])



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Levenberg-Marquardt algorithm in JAX."""

from typing import Any
from typing import Callable
from typing import Literal
from typing import NamedTuple
from typing import Optional
from typing import Union

from dataclasses import dataclass
import warnings

import jax
import jax.numpy as jnp
import jax.scipy as jsp

from nemos.third_party.jaxopt.jaxopt._src import base
from nemos.third_party.jaxopt.jaxopt._src.linear_solve import solve_cg
from nemos.third_party.jaxopt.jaxopt._src.linear_solve import solve_cholesky
from nemos.third_party.jaxopt.jaxopt._src.linear_solve import solve_inv
from nemos.third_party.jaxopt.jaxopt._src.linear_solve import solve_lu
from nemos.third_party.jaxopt.jaxopt._src.linear_solve import solve_qr
from nemos.third_party.jaxopt.jaxopt._src.tree_util import tree_l2_norm, tree_inf_norm


class LevenbergMarquardtState(NamedTuple):
  """Named tuple containing state information."""
  iter_num: int
  damping_factor: float
  increase_factor: float
  residual: Any
  value: Any
  delta: Any
  error: float
  gradient: Any
  jac: Any
  jt: Any
  jtj: Any
  hess_res: Any
  aux: Optional[Any] = None

@dataclass(eq=False)
class LevenbergMarquardt(base.IterativeSolver):
  """Levenberg-Marquardt (LM) nonlinear least-squares solver.

    Given the residual function `func` (x): R^n -> R^m, `least_squares` finds a
    local minimum of the cost function F(x):

    ```
    argmin_x F(x) = 0.5 * sum(f_i(x)**2), i = 0, ..., m - 1
    f(x) = func(x, *args)
    ```

    This results in solving the following normal equation:
    ```
    (J^T @ J + µ I) @ ∆params = -J^T @ f(x)
    ```
    we use this method as recommended by K. Madsen & H. B. Nielsen in the book
    "Introduction to Optimization and Data Fitting" page 122.


    If stop_criterion is 'madsen-nielsen', the convergence is achieved once the
    coeff update satisfies ``||dcoeffs||_2 <= xtol * (||coeffs||_2 + xtol) `` or
    the gradient satisfies ``||grad(f)||_inf <= gtol``.

    It is possible to find the LM step by solving the following equation:
    ```
    [     J    ]              [ -f(x) ]
    |          | @ ∆params =  |       |
    [ Sqrt(µ)I ]              [   0   ]
    ```
    If LHS matrix is multiplied by its transpose, one can show that solving this
    equation is the same as the normal equation. This is the approach we take if
    QR factorization method is chosen. Note that since we are not multipyling the
    LHS matrix by its transpose, the condition number of the LHS matrix is more
    accurate and therefore this method can be used for ill-conditioned Jacobians.
  Attributes:
    residual_fun: a smooth function of the form ``residual_fun(x, *args,
      **kwargs)``.
    maxiter: maximum increase_factormber of iterations.
    damping_parameter: The parameter which adds a correction to the equation
      derived for updating the coefficients using Gauss-Newton method. Please
      see section 3.2. of K. Madsen et al. in the book "Methods for nonlinear
      least squares problems" for more information.
    stop_criterion: The criterion to use for the convergence of the while loop.
      e.g., for 'madsen-nielsen' the criteria is to satisfy the two equations
      for delta_params and gradient that is mentioned above. If 'grad-l2' is
      selected, the convergence is achieved if l2 of gradient is smaller or
      equal to tol.
    tol: tolerance.
    xtol: float, optional The convergence tolerance for the second norm of the
      coefficient update.
    gtol: float, optional The convergence tolerance for the inf norm of the
      residual gradient.
    solver: str, optional The solver to use when finding delta_params, the
      update to the params in each iteration. This is done through solving a
      system of linear equation Ax=b. 'cholesky', 'lu', or 'qr' factorizations,
      'inv' (explicit multiplication with matrix inverse). Note that the inverse
      approach is the most expensive and least accurate and is just given as an
      option for legacy reasons. The user can provide custom solvers, for example
      using jaxopt.linear_solve.solve_cg which are more scalable for runtime but
      take longer compilations. Cholesky is faster than inverse since it uses
      the symmetry feature of A. QR factorization is preferred for ill-posed
      stiff problems due to not solving the normal equation.
    geodesic: bool, if we would like to include the geodesic acceleration when
      solving for the delta_params in every iteration.
    contribution_ratio_threshold: float, the threshold for acceleration/velocity
      ratio. We update the parameters in the algorithm only if the ratio is
      smaller than this threshold value.
    verbose: bool, whether to print information on every iteration or not.
    jac_fun: Callable, a function to calculate the Jacobian. If not None, this
      function is used instead of directly calculating it using ``jax.jacfwd``.
    materialize_jac: bool, whether to materialize Jacobian. If this option is
      True, Jacobian is either calculated using ``jax.jacfwd`` or obtained from
      ``jac_fun`` and all variables depending on it, such as J^T or J^T.J are
      obtained directly. If False, all of the Jacobian dependent variables are
      indirectly obtained using operators on the basis of ``jax.jvp`` and
      ``jax.vjp``.
    implicit_diff: bool, whether to enable implicit diff or autodiff of unrolled
      iterations.
    implicit_diff_solve: the linear system solver to use.
    has_aux: whether ``residual_fun`` outputs auxiliary data or not.
    jit: whether to JIT-compile the bisection loop (default: True).
    unroll: whether to unroll the bisection loop (default: "auto").

  Reference: This algorithm is for finding the best fit parameters based on the
    algorithm 6.18 provided by K. Madsen & H. B. Nielsen in the book
    "Introduction to Optimization and Data Fitting".
  """
  residual_fun: Callable

  maxiter: int = 30

  damping_parameter: float = 1e-6
  stop_criterion: Literal['grad-l2-norm', 'madsen-nielsen'] = 'grad-l2-norm'
  tol: float = 1e-3
  xtol: float = 1e-3
  gtol: float = 1e-3

  solver: Union[Literal['cholesky', 'lu', 'qr', 'inv'], Callable] = solve_cg
  geodesic: bool = False
  contribution_ratio_threshold = 0.75

  verbose: Union[bool, int] = False
  jac_fun: Optional[Callable[..., jnp.ndarray]] = None
  materialize_jac: bool = False
  implicit_diff: bool = True
  implicit_diff_solve: Optional[Callable] = None
  has_aux: bool = False
  jit: bool = True
  unroll: base.AutoOrBoolean = 'auto'

  # We are overriding the _cond_fun of the base solver to enable stopping based
  # on gradient or delta_params
  def _cond_fun(self, inputs):
    params, state = inputs[0]
    if self.verbose:
      print_iteration(state)
    if self.stop_criterion == 'madsen-nielsen':
      tree_mul_term = self.xtol * (tree_l2_norm(params) - self.xtol)
      return jnp.all(jnp.array([
        tree_inf_norm(state.gradient) > self.gtol,
        tree_l2_norm(state.delta) > tree_mul_term
      ]))
    elif self.stop_criterion == 'grad-l2-norm':
      return state.error > self.tol
    else:
      raise NotImplementedError

  def init_state(self, init_params: Any, *args,
                 **kwargs) -> LevenbergMarquardtState:
    """Initialize the solver state.

    Args:
      init_params: pytree containing the initial parameters.
      *args: additional positional arguments to be passed to ``residual_fun``.
      **kwargs: additional keyword arguments to be passed to ``residual_fun``.

    Returns:
      state
    """
    # Compute actual values of state variables at init_param
    residual, aux = self._fun_with_aux(init_params, *args, **kwargs)

    if self.materialize_jac:
      jac = self._jac_fun(init_params, *args, **kwargs)
      jt = jac.T
      jtj = jt @ jac
      gradient = jt @ residual
      damping_factor = self.damping_parameter * jnp.max(jnp.diag(jtj))
      if self.geodesic:
        hess_res = self._hess_res_fun(init_params, *args, **kwargs)
      else:
        hess_res = None
    else:
      jac = None
      jt = None
      jtj = None
      hess_res = None
      gradient = self._jt_op(init_params, residual, *args, **kwargs)
      jtj_diag = self._jtj_diag_op(init_params, *args, **kwargs)
      damping_factor = self.damping_parameter * jnp.max(jtj_diag)

    delta_params = jnp.zeros_like(init_params)

    return LevenbergMarquardtState(
        iter_num=jnp.asarray(0),
        damping_factor=damping_factor,
        increase_factor=2,
        error=tree_l2_norm(gradient),
        residual=residual,
        value=0.5 * jnp.sum(jnp.square(residual)),
        delta=delta_params,
        gradient=gradient,
        jac=jac,
        jt=jt,
        jtj=jtj,
        hess_res=hess_res,
        aux=aux)

  def update_state_using_gain_ratio(self, gain_ratio, contribution_ratio_diff,
                                    gain_ratio_test_init_state, *args,
                                    **kwargs):
    """The function to return state variables based on gain ratio.
      Please see by page 120-121 of the book "Introduction to Optimization and
      Data Fitting" by K. Madsen & H. B. Nielsen for details.
    """

    def gain_ratio_test_true_func(params, damping_factor,
                                  increase_factor, residual, gradient, jac, jt, jtj,
                                  hess_res, updated_params, aux):

      params = updated_params

      residual, aux = self._fun_with_aux(params, *args, **kwargs)

      # Calculate gradient based on Eq. 6.6 of "Introduction to optimization
      # and data fitting" g=JT * r, where J is jacobian and r is residual.
      # TODO: QR factorization solver doesn't require jt and jtj. Can be
      # skipped for increasing efficiency, if they are not required by user.
      if self.materialize_jac:
        # Calculate Jacobian and it's transpose based on the updated coeffs.
        jac = self._jac_fun(params, *args, **kwargs)
        jt = jac.T
        #  J^T.J is the gauss newton approximate hessian.
        jtj = jt @ jac
        gradient = jt @ residual
        if self.geodesic:
          hess_res = self._hess_res_fun(params, *args, **kwargs)
        else:
          hess_res = None
      else:
        jac = None
        jt = None
        jtj = None
        hess_res = None
        gradient = self._jt_op(params, residual, *args, **kwargs)

      damping_factor = damping_factor * jax.lax.max(1 / 3, 1 -
                                                    (2 * gain_ratio - 1)**3)
      increase_factor = 2

      return params, damping_factor, increase_factor, residual, gradient, jac, jt, jtj, hess_res, aux

    def gain_ratio_test_false_func(params, damping_factor,
                                   increase_factor, residual, gradient, jac, jt, jtj,
                                   hess_res, updated_params, aux):
      damping_factor = jnp.minimum(damping_factor * increase_factor, self.damping_factor_max)
      increase_factor = jnp.minimum(2 * increase_factor, self.increase_factor_max)
      return params, damping_factor, increase_factor, residual, gradient, jac, jt, jtj, hess_res, aux

    # Calling the jax condition function:
    # Note that only the parameters that are used in the rest of the program
    # are returned by this function.

    gain_ratio_test_is_met = jnp.logical_and(gain_ratio > 0.0,
                                             contribution_ratio_diff <= 0.0)

    gain_ratio_test_is_met_ret = gain_ratio_test_true_func(
        *gain_ratio_test_init_state)
    gain_ratio_test_not_met_ret = gain_ratio_test_false_func(
        *gain_ratio_test_init_state)

    gain_ratio_test_is_met_ret = jax.tree.map(
        lambda x: gain_ratio_test_is_met * x, gain_ratio_test_is_met_ret)

    gain_ratio_test_not_met_ret = jax.tree.map(
        lambda x: (1.0 - gain_ratio_test_is_met) * x,
        gain_ratio_test_not_met_ret)

    params, damping_factor, increase_factor, residual, gradient, jac, jt, jtj, hess_res, aux = jax.tree.map(
        lambda x, y: x + y, gain_ratio_test_is_met_ret,
        gain_ratio_test_not_met_ret)

    return params, damping_factor, increase_factor, residual, gradient, jac, jt, jtj, hess_res, aux

  def update_state_using_delta_params(self, loss_curr, params, delta_params,
                                      contribution_ratio_diff, damping_factor,
                                      increase_factor, residual, gradient, jac,
                                      jt, jtj, hess_res, aux, *args, **kwargs):
    """The function to return state variables based on delta_params.

    Define the functions required for the major conditional of the algorithm,
    which checks the magnitude of dparams and checks if it is small enough.
    for the value of dparams.
    """

    updated_params = params + delta_params

    residual_next = self._fun(updated_params, *args, **kwargs)

    # Calculate denominator of the gain ratio based on Eq. 6.16, "Introduction
    # to optimization and data fitting", L(0)-L(hlm)=0.5*hlm^T*(mu*hlm-g).
    gain_ratio_denom = 0.5 * delta_params.T @ (
        damping_factor * delta_params - gradient)

    # Current value of loss function F=0.5*||f||^2.
    loss_next = 0.5 * jnp.sum(jnp.square(residual_next))

    gain_ratio = (loss_curr - loss_next) / gain_ratio_denom

    gain_ratio_test_init_state = (params, damping_factor, increase_factor,
                                  residual, gradient, jac, jt, jtj, hess_res,
                                  updated_params, aux)

    # Calling the jax condition function:
    # Note that only the parameters that are used in the rest of the program
    # are returned by this function.

    params, damping_factor, increase_factor, residual, gradient, jac, jt, jtj, hess_res, aux = (
        self.update_state_using_gain_ratio(gain_ratio, contribution_ratio_diff,
                                           gain_ratio_test_init_state, *args,
                                           **kwargs))

    return params, damping_factor, increase_factor, residual, gradient, jac, jt, jtj, hess_res, aux

  def update(self, params, state: NamedTuple, *args, **kwargs) -> base.OptStep:
    """Performs one iteration of the least-squares solver.

    Args:
      params: pytree containing the parameters.
      state: named tuple containing the solver state.

    Returns:
      (params, state)
    """

    # Current value of the loss function F=0.5*||f||^2.
    loss_curr = state.value

    # TODO: clean up the linear_solver to take matrix directly and then clean up
    # linear equation solves below for materialize_jac=True to call them.

    # For geodesic acceleration, we calculate  jtrpp=JT * r",
    # where J is jacobian and r" is second order directional derivative.
    if self.materialize_jac:
      # Note that instead of taking the inverse of jtj_corr and multiply that
      # by state.gradient, we prefer to use `jsp.solve`, which uses Cholesky of
      # jtj_corr and uses that to obtain velocity. This has the advantage of
      # lower number of floating point operations and therefore less numerical
      # error which can be helpful for the case of single precision arithmatics.
      if self.solver == 'qr':
        damping_term = jnp.sqrt(state.damping_factor) * jnp.identity(params.size)
        aug_jac = jnp.vstack([state.jac, damping_term])
        aug_res = jnp.hstack([state.residual, jnp.zeros(params.size)])
        q, r = jsp.linalg.qr(aug_jac, mode='economic')
        velocity = jsp.linalg.solve_triangular(r, q.T @ aug_res)
      else:
        damping_term = state.damping_factor * jnp.identity(params.size)
        jtj_corr = state.jtj + damping_term
        if self.solver == 'lu':
          velocity = jnp.linalg.solve(jtj_corr, state.gradient)
        elif self.solver == 'inv':
          jtj_corr_inv = jnp.linalg.inv(jtj_corr)
          velocity = jnp.dot(jtj_corr_inv, state.gradient)
        else:
          velocity = jsp.linalg.solve(jtj_corr, state.gradient, assume_a='pos')
      delta_params = velocity
      if self.geodesic:
        rpp = (state.hess_res @ velocity) @ velocity
        # Note the same as above here that we could use inverse of jtj_corr but
        # chose to use solve for higher performance and lower numerical error.
        if self.solver == 'qr':
          aug_rpp = jnp.hstack([rpp, jnp.zeros(params.size)])
          acceleration = jsp.linalg.solve_triangular(r, q.T @ aug_rpp)
        elif self.solver == 'lu':
          acceleration = jnp.linalg.solve(jtj_corr, state.jt) @ rpp
        elif self.solver == 'inv':
          acceleration = jnp.dot(jtj_corr_inv, state.jt) @ rpp
        else:
          acceleration = jsp.linalg.solve(jtj_corr, state.jt, assume_a='pos') @ rpp

        delta_params += 0.5*acceleration
    else:
      matvec = lambda v: self._jtj_op(params, v, *args, **kwargs)

      velocity, acceleration, delta_params = self._solve_linear_eqs(
        matvec, state, params, *args, **kwargs
      )

    if self.geodesic:
      contribution_ratio_diff = jnp.linalg.norm(acceleration) / jnp.linalg.norm(
          velocity) - self.contribution_ratio_threshold
    else:
      contribution_ratio_diff = 0.0

    # Negative coefficient is due to the sign of the RHS vector in the update equation
    # (J^T @ J + µ I) @ ∆params = -J^T @ f(x).
    delta_params = -delta_params

    # Checking if the dparams satisfy the "sufficiently small" criteria.
    params, damping_factor, increase_factor, residual, gradient, jac, jt, jtj, hess_res, aux = (
        self.update_state_using_delta_params(loss_curr, params, delta_params,
                                             contribution_ratio_diff,
                                             state.damping_factor,
                                             state.increase_factor,
                                             state.residual, state.gradient,
                                             state.jac, state.jt, state.jtj,
                                             state.hess_res, state.aux, *args, **kwargs))

    new_value = 0.5 * jnp.sum(jnp.square(residual))
    state = LevenbergMarquardtState(
        iter_num=state.iter_num + 1,
        damping_factor=damping_factor,
        increase_factor=increase_factor,
        error=tree_l2_norm(gradient),
        residual=residual,
        value=new_value,
        delta=delta_params,
        gradient=gradient,
        jac=jac,
        jt=jt,
        jtj=jtj,
        hess_res=hess_res,
        aux=aux)

    if self.verbose:
      self.log_info(
          state,
          error_name="Gradient Norm",
          additional_info={
              "Objective Value": new_value,
              "Damping Factor": damping_factor
          }
      )
    return base.OptStep(params=params, state=state)

  def __post_init__(self):
    super().__post_init__()

    if isinstance(self.solver, Callable):
      self.solver_fn = self.solver
    elif self.solver == 'cholesky':
      self.solver_fn = solve_cholesky
    elif self.solver == 'lu':
      self.solver_fn = solve_lu
    elif self.solver == 'qr':
      self.solver_fn = solve_qr
    elif self.solver == 'inv':
      self.solver_fn = solve_inv
    else:
      raise NotImplementedError
    if self.has_aux:
      self._fun_with_aux = self.residual_fun
      self._fun = lambda *a, **kw: self._fun_with_aux(*a, **kw)[0]
    else:
      self._fun = self.residual_fun
      self._fun_with_aux = lambda *a, **kw: (self.residual_fun(*a, **kw), None)

    # Define maximum value for the damping factor and increase factor. This is
    # particularly helpful for X32 usage.
    self.damping_factor_max = jnp.asarray(2**32, dtype=float)
    self.increase_factor_max = jnp.asarray(2**32, dtype=float)

    # For geodesic acceleration, we define Hessian of the residual function.
    if self.materialize_jac:
      if self.jac_fun is None:
        self._jac_fun = jax.jacfwd(self._fun, argnums=(0))
        if self.geodesic:
          self._hess_res_fun = jax.jacfwd(
              jax.jacfwd(self._fun, argnums=(0)), argnums=(0))
    else:
      if self.jac_fun:
        self._jac_fun = self.jac_fun
        if self.geodesic:
          self._hess_res_fun = jax.jacfwd(self.jac_fun, argnums=(0))
      if self.solver == 'qr':
        raise ValueError("QR factorization solver materializes Jacobian to solve the augmented "
                         "Jacobian equation instead of the normal equation. Hence, this solver "
                         "choice is inconsistent with materialize_jac=False.")
      if self.solver in ['lu', 'cholesky', 'inv']:
        warnings.warn(f"The linear solver {self.solver} that requires materialization of "
                      "J^T.J matrix is used with materialize_jac=False, which may cause a "
                      "computational overhead. Consider using either a matrix-free iterative "
                      "solver such as cg or bicg or using materialize_jac=True.", category=UserWarning)

  def optimality_fun(self, params, *args, **kwargs):
    """Optimality function mapping compatible with ``@custom_root``."""
    residual = self._fun(params, *args, **kwargs)
    return self._jt_op(params, residual, *args, **kwargs)

  def _jt_op(self, params, residual, *args, **kwargs):
    """Product of J^T and residual -- J: jacobian of fun at params."""
    fun_with_args = lambda p: self._fun(p, *args, **kwargs)
    _, vjpfun = jax.vjp(fun_with_args, params)
    jt_op_val, = vjpfun(residual)
    return jt_op_val

  def _jtj_op(self, params, vec, *args, **kwargs):
    """Product of J^T.J with vec using vjp & jvp, where J is jacobian of fun at params."""
    fun_with_args = lambda p: self._fun(p, *args, **kwargs)
    _, vjpfun = jax.vjp(fun_with_args, params)
    _, jvp_val = jax.jvp(fun_with_args, (params,), (vec,))
    jtj_op_val, = vjpfun(jvp_val)
    return jtj_op_val

  def _jtj_diag_op(self, params, *args, **kwargs):
    """Diagonal elements of J^T.J, where J is jacobian of fun at params."""
    diag_op = lambda v: v.T @ self._jtj_op(params, v, *args, **kwargs)
    return jax.vmap(diag_op)(jnp.eye(len(params))).T

  def _d2fvv_op(self, primals, tangents1, tangents2, *args, **kwargs):
    """Product with d2f.v1v2."""
    fun_with_args = lambda p: self._fun(p, *args, **kwargs)
    g = lambda pr: jax.jvp(fun_with_args, (pr,), (tangents1,))[1]
    return jax.jvp(g, (primals,), (tangents2,))[1]

  def _solve_linear_eqs(self, matvec, state, params, *args, **kwargs):
    """Solves the linear equations to obtain velocity and acceleration."""
    if isinstance(self.solver, Callable):
      solver_options = {'ridge': state.damping_factor, 'init': state.delta}
    else:
      solver_options = {'ridge': state.damping_factor}
    velocity = self.solver_fn(matvec, state.gradient, **solver_options)
    delta_params = velocity
    if self.geodesic:
      rpp = self._d2fvv_op(params, velocity, velocity, *args, **kwargs)
      jtrpp = self._jt_op(params, rpp, *args, **kwargs)
      acceleration = self.solver_fn(matvec, jtrpp, ridge=state.damping_factor)
      delta_params += 0.5*acceleration
    else:
      acceleration = jnp.zeros_like(velocity)

    return (velocity, acceleration, delta_params)


def print_iteration(state: LevenbergMarquardtState):
  jax.debug.print("Iteration: {iter}, Value: {value}, ||Gradient||: {error}, Damping Factor: {damp}",
                iter=state.iter_num, value=state.value, error=state.error, damp=state.damping_factor)



"""Custom Basis Class.

Facilitate the construction of a custom basis class.
"""

from __future__ import annotations

import inspect
import itertools
import re
from copy import deepcopy
from numbers import Number
from typing import TYPE_CHECKING, Callable, Iterable, List, Optional, Tuple

import numpy as np
import pynapple as nap
from numpy.typing import ArrayLike, NDArray
from pynapple import Tsd, TsdFrame, TsdTensor

from nemos.typing import FeatureMatrix

from ..base_class import Base
from ..type_casting import support_pynapple
from ..utils import format_repr
from . import AdditiveBasis, MultiplicativeBasis
from ._basis_mixin import BasisMixin, BasisTransformerMixin, set_input_shape_state
from ._check_basis import _check_transform_input
from ._composition_utils import (
    _check_unique_shapes,
    _check_valid_shape_tuple,
    add_docstring,
    count_positional_and_var_args,
    infer_input_dimensionality,
    is_basis_like,
    multiply_basis_by_integer,
    promote_to_transformer,
    raise_basis_to_power,
    set_input_shape,
)

if TYPE_CHECKING:
    from . import TransformerBasis


def simplify_func_repr(string: str):
    """
    Simplify function repr.

    Simplify function repr by dropping the address and replace "functools.partial" with "partial".

    Parameters
    ----------
    string:
        The string repr of a parameter.

    Returns
    -------
        A simplified version of the string.

    """
    pattern = r"<function (.+) at 0x[0-9a-fA-F]+>"
    new_string = deepcopy(string)
    for match in re.finditer(pattern, string):
        func_name = match.group(1)
        orig_repr = match.group(0)
        new_string = re.sub(orig_repr, func_name, new_string, count=1)
    new_string = re.sub(r"functools\.partial", "partial", new_string)
    return new_string


class FunctionList:
    def __init__(self, funcs, name="b"):
        self.funcs_ = list(funcs)  # store as list
        self.name = name  # optional label prefix

    def __getitem__(self, idx):
        return self.funcs_[idx]

    def __len__(self):
        return len(self.funcs_)

    def __iter__(self):
        return iter(self.funcs_)

    @staticmethod
    def _unwrap(func):
        return func.__wrapped__ if hasattr(func, "__wrapped__") else func

    def __repr__(self):
        unwrap = [self._unwrap(f) for f in self.funcs_]
        if len(unwrap) <= 2:
            return simplify_func_repr(repr(unwrap))
        return f"[{simplify_func_repr(repr(unwrap[0]))}, ..., {simplify_func_repr(repr(unwrap[-1]))}]"


def apply_f_vectorized(
    func: Callable[[NDArray], NDArray], *xi: NDArray, ndim_input: int = 1, **kwargs
):
    """Iterate over the output dim and apply the function to all input combination."""

    # check if no dimension needs vectorization
    if all(x.ndim == ndim_input for x in xi):
        return func(*xi, **kwargs)[..., np.newaxis]

    # Get the vectorized shape (should be the same for all inputs)
    vec_shape = xi[0].shape[ndim_input:]

    # Generate all combinations of vectorized indices in the correct order
    vec_indices = itertools.product(*[range(dim) for dim in vec_shape])

    # Collect results for each vectorized index combination
    results = []
    for indices in vec_indices:
        # Extract slices for this combination of indices
        slices = [x[(slice(None),) * ndim_input + indices] for x in xi]

        # Apply function to the slices
        result = func(*slices, **kwargs)
        results.append(result[..., np.newaxis])

    # Concatenate along the last axis
    return np.concatenate(results, axis=-1)


def check_valid_shape(shape):
    is_numeric = not all(isinstance(i, Number) for i in shape)
    is_int = all(i == int(i) for i in shape) if is_numeric else False
    if not (is_numeric and is_int):
        raise ValueError("`output_shape` must be an iterable of integers.")


class CustomBasis(BasisMixin, BasisTransformerMixin, Base):
    """
    Custom basis class.

    Create a custom basis class from a list of callables (the basis functions).

    Parameters
    ----------
    funcs:
        List of basis functions.
    ndim_input:
        Dimensionality of the input for each sample, i.e. if your time series is of shape ``(n_samples, n, m)``,
        ``ndim_input`` is two.
    output_shape:
        Shape of the output excluding the number of samples. Set automatically when `compute_features` is called.
    basis_kwargs:
        Additional keyword arguments to pass to the basis function.
    pynapple_support:
        Enable pynapple support if True.
    label:
        The label of the basis function.
    is_complex : bool, optional
        Whether the basis should be treated as complex. This flag ensures that
        multiplication with other bases behaves correctly: two real bases, or a real
        and a complex basis, can be multiplied, but two complex bases cannot. This
        restriction exists because after multiplication, ``basis.compute_features``
        does not distinguish between real and imaginary components, which would lead
        to incorrect outputs.

    Examples
    --------
    >>> import numpy as np
    >>> import nemos as nmo
    >>> from functools import partial
    >>> # Define a function
    >>> def decay_exp(x, rate, shift=0):
    ...     return np.exp(-rate * (x + shift)**2)
    >>> # Define a list of basis functions
    >>> funcs = [partial(decay_exp, rate=r) for r in np.linspace(0, 1, 10)]
    >>> bas = nmo.basis.CustomBasis(funcs=funcs, basis_kwargs=dict(shift=1))
    >>> bas
    CustomBasis(
        funcs=[partial(decay_exp, rate=np.float64(0.0)), ..., partial(decay_exp, rate=np.float64(1.0))],
        ndim_input=1,
        basis_kwargs={'shift': 1},
        pynapple_support=True,
        is_complex=False
    )
    >>> samples = np.linspace(0, 1, 50)
    >>> X = bas.compute_features(samples)
    >>> X.shape
    (50, 10)
    >>> # Can be composed with other basis (including other custom basis)
    >>> add = bas + bas
    >>> X = add.compute_features(samples, samples)
    >>> X.shape
    (50, 20)
    """

    def __init__(
        self,
        funcs: List[Callable[[NDArray], NDArray]] | Callable[[NDArray], NDArray],
        ndim_input: int = 1,
        output_shape: Optional[Tuple[int, ...] | int] = None,
        basis_kwargs: Optional[dict] = None,
        pynapple_support: bool = True,
        label: Optional[str] = None,
        is_complex: bool = False,
    ):
        self._pynapple_support = bool(pynapple_support)
        self.funcs = funcs
        self.ndim_input = int(ndim_input)

        if output_shape is None:
            output_shape = ()

        self.output_shape = output_shape

        self._input_shape_product = None

        self._n_input_dimensionality = infer_input_dimensionality(self)
        self._n_basis_funcs = len(self.funcs)

        self.basis_kwargs = basis_kwargs
        self._is_complex = bool(is_complex)
        super().__init__(label=label)

    @property
    def is_complex(self):
        # custom classes could be complex or real, so the attribute
        # is an instance attribute not a class attribute
        return self._is_complex

    @property
    def pynapple_support(self) -> bool:
        """Support pynapple Tsd/TsdFrame/TsdTensor as inputs."""
        return self._pynapple_support

    @property
    def funcs(self):
        """User defined list of basis functions."""
        return self._funcs

    @funcs.setter
    def funcs(self, val: Iterable[Callable[[NDArray, ...], NDArray]]):
        if isinstance(val, Callable):
            val = [val]
        val = FunctionList(val)

        if not all(isinstance(f, Callable) for f in val):
            raise ValueError("User must provide an iterable of callable.")

        if hasattr(self, "_n_input_dimensionality"):
            inp_dim = sum(count_positional_and_var_args(f)[0] for f in val)
            if inp_dim != self._n_input_dimensionality:
                raise ValueError(
                    "The number of input time series required by the CustomBasis must be consistent. "
                    "Redefine a CustomBasis for a different number of inputs."
                )
        self._funcs = val

    @property
    def n_basis_funcs(self) -> int:
        """The number of basis."""
        return self._n_basis_funcs

    @property
    def output_shape(self) -> Tuple[int, ...]:
        """The shape of the output excluding the number of samples and the number of basis functions."""
        return self._output_shape

    @output_shape.setter
    def output_shape(self, output_shape: Tuple[int] | int):
        if isinstance(output_shape, int):
            if output_shape == 1:
                self._output_shape = ()
            elif output_shape > 1:
                self._output_shape = (output_shape,)
            else:
                raise ValueError(
                    f"Output shape must be strictly positive (> 0), {output_shape} provided instead."
                )
        else:
            try:
                output_shape = tuple(output_shape)
            except TypeError:
                raise TypeError(
                    "`output_shape` must be an iterable of positive integers or a positive integer."
                )
            _check_valid_shape_tuple(output_shape)
            self._output_shape = output_shape

    @property
    def basis_kwargs(self) -> dict:
        """Additional keyword arguments to pass to the basis functions."""
        return self._basis_kwargs

    @basis_kwargs.setter
    def basis_kwargs(self, basis_kwargs: dict):
        # store args and kwargs
        basis_kwargs = basis_kwargs if basis_kwargs is not None else {}
        if not isinstance(basis_kwargs, dict):
            raise ValueError("`basis_kwargs` must be a dictionary.")
        sig = inspect.signature(apply_f_vectorized)
        invalid_kwargs = {
            p.name
            for p in sig.parameters.values()
            if p.kind is inspect.Parameter.KEYWORD_ONLY
        }
        if invalid_kwargs.intersection(basis_kwargs.keys()):
            raise ValueError(
                f"Invalid kwargs name in ``basis_kwargs``: {invalid_kwargs.intersection(basis_kwargs.keys())}. "
                "Change parameter name in your function definition."
            )
        self._basis_kwargs = basis_kwargs

    def compute_features(
        self, *xi: ArrayLike | Tsd | TsdFrame | TsdTensor
    ) -> FeatureMatrix:
        """
        Apply the basis transformation to the input data.

        This method applies each function in ``self.funcs`` to the input arrays ``*xi``.
        These functions are called with the arguments ``(*xi, **self.basis_kwargs)`` and must return
        an array of shape ``(n_samples, ...)``, where the first dimension corresponds to the number of samples,
        and the output must have at least one dimension (i.e., ``ndim >= 1``).

        The outputs of all function calls are reshaped into 2D arrays with shape ``(n_samples, n_output)``, and
        then concatenated along the feature axis (second dimension) to form the full design matrix.

        If the input arrays have more dimensions than ``self.ndim_input``, the function calls are automatically
        vectorized over the additional axes. This is done using Python loops, which may be slow. For better
        performance, users are encouraged to provide fully vectorized functions.

        Parameters
        ----------
        *xi :
            Input arrays. Each must have at least ``self.ndim_input`` dimensions. If extra dimensions are present,
            they are interpreted as batch or window dimensions over which the basis functions are applied.

        Returns
        -------
        :
            The resulting design matrix, with one row per sample and one column per output feature.

        Examples
        --------
        >>> import nemos as nmo
        >>> import numpy as np
        >>> from functools import partial
        >>> def power_func(n, x):
        ...     return x ** n
        >>> bas = nmo.basis.CustomBasis([partial(power_func, 1), partial(power_func, 2)])
        >>> bas.compute_features(np.arange(1, 4))
        array([[1., 1.],
               [2., 4.],
               [3., 9.]])
        """
        xi = _check_transform_input(self, *xi)
        if any(x.ndim < self.ndim_input for x in xi):
            invalid_dims = [x.ndim for x in xi if x.ndim < self.ndim_input]
            raise ValueError(
                f"Each input must have at least {self.ndim_input} dimensions, as required by this basis. "
                f"However, some inputs have fewer dimensions: {invalid_dims}."
            )
        _check_unique_shapes(*xi, basis=self)
        set_input_shape(self, *xi)
        design_matrix = self.evaluate(
            *xi
        )  # (n_samples, *n_output_shape, n_vec_dim, n_basis)
        # return a model design
        return design_matrix.reshape((xi[0].shape[0], -1))

    def evaluate(self, *xi: NDArray):
        """
        Evaluate the basis functions in a vectorized form at the given sample points.

        Parameters
        ----------
        *xi :
            The samples at which the basis functions are evaluated. Each element in `xi` corresponds
            to an input dimension, and must be broadcastable to a common shape along the sample axis.
            The shape of each input array should be (n_samples, ...) where the first axis indexes samples.

        Returns
        -------
        basis_funcs :
            The basis functions evaluated at the given input points, with shape
            (n_samples, n_vect_input * n_basis_funcs), n_vect_input is the number of inputs that are
            vectorized.

        Notes
        -----
        This method supports both NumPy and pynapple inputs. If pynapple support is enabled,
        the inputs and outputs are automatically cast using the configured backend (e.g., JAX or NumPy).
        Evaluation is performed by applying a vectorized function over each basis function and
        concatenating the results along the last axis.

        Examples
        --------
        >>> import numpy as np
        >>> from nemos.basis import CustomBasis
        >>> basis = CustomBasis(funcs=[lambda x: x, lambda x: x**2])
        >>> x = np.linspace(0, 1, 10)
        >>> out = basis.evaluate(x)
        >>> out.shape
        (10, 2)
        >>> # vectorize over 3 inputs
        >>> out = basis.evaluate(np.random.randn(10, 3))
        >>> out.shape
        (10, 3, 2)

        """
        if self._pynapple_support:
            conv_type = "numpy" if nap.nap_config.backend == "numba" else "jax"
            apply_func = support_pynapple(conv_type)(apply_f_vectorized)
        else:
            apply_func = apply_f_vectorized

        # Get individual function results
        func_results = [
            apply_func(f, *xi, **self.basis_kwargs, ndim_input=self.ndim_input)
            for f in self.funcs
        ]

        # Stack functions first, then reorder
        stacked = np.stack(
            func_results, axis=-1
        )  # (n_samples, *out_shape, n_vec_features, n_funcs)
        self.output_shape = stacked.shape[1:-2]

        # no vectorization
        if all(x.ndim == self.ndim_input for x in xi):
            stacked = stacked[..., 0, :]
        return stacked

    @set_input_shape_state(states=("_input_shape_product", "_input_shape_", "_label"))
    def __sklearn_clone__(self) -> "CustomBasis":
        """Clone the basis while preserving attributes related to input shapes.

        This method ensures that input shape attributes (e.g., `_input_shape_product`,
        `_input_shape_`) are preserved during cloning. Reinitializing the class
        as in the regular sklearn clone would drop these attributes, rendering
        cross-validation unusable.
        """
        klass = self.__class__(**self.get_params())
        return klass

    @property
    def n_output_features(self):
        """The number of output features, i.e. the number of columns of the design matrix."""
        if self.input_shape is None:
            return None
        # Computation for number of output features:
        # 1. Compute the number of vectorized dimensions:
        #    - discard axis corresponding to input dimensionality
        #    - multiply the shape of the remaining axis
        # 2. multiply by the number of basis and the shape of the output

        # Note that self._input_shape_ could be:
        #   1. A list of tuple, if the funcs require > 1 input.
        #   2. A single tuple, if the funcs require 1 input.
        ishape = (
            [self._input_shape_]
            if not isinstance(self._input_shape_, list)
            else self._input_shape_
        )
        vec_inp = np.prod([p for shape in ishape for p in shape[self.ndim_input - 1 :]])
        return int(vec_inp * np.prod(self.output_shape) * len(self.funcs))

    @staticmethod
    def _reshape_concatenated_arrays(
        array: NDArray, bas: "CustomBasis", axis: int
    ) -> NDArray:
        # reshape the arrays to match input shapes
        shape = list(array.shape)
        array = array.reshape(
            shape[:axis]
            + [
                *bas.output_shape,
                *(
                    i
                    for shape in bas._input_shape_
                    for i in shape[bas.ndim_input - 1 :]
                ),
                len(bas.funcs),
            ]
            + shape[axis + 1 :]
        )
        return array

    @promote_to_transformer
    def __add__(self, other: BasisMixin) -> AdditiveBasis:
        return AdditiveBasis(self, other)

    @promote_to_transformer
    def __rmul__(self, other: BasisMixin | int) -> BasisMixin:
        return self.__mul__(other)

    @promote_to_transformer
    def __mul__(self, other: BasisMixin | int) -> BasisMixin:
        if isinstance(other, int):
            return multiply_basis_by_integer(self, other)

        if not is_basis_like(other):
            raise TypeError(
                "Basis multiplicative factor should be a Basis object or a positive integer!"
            )

        return MultiplicativeBasis(self, other)

    @promote_to_transformer
    def __pow__(self, exponent) -> BasisMixin:
        return raise_basis_to_power(self, exponent)

    def __repr__(self, n=0):
        rep = format_repr(self, multiline=True)
        tab = "    "
        return rep.replace("\n", f"\n{tab * n}")

    def __len__(self) -> int:
        return 1

    @add_docstring("split_by_feature", BasisMixin)
    def split_by_feature(
        self,
        x: NDArray,
        axis: int = 1,
    ):
        """
        Examples
        --------
        >>> import nemos as nmo
        >>> import numpy as np
        >>> from functools import partial
        >>> def power_func(n, x):
        ...     return x ** n
        >>> bas = nmo.basis.CustomBasis([partial(power_func, 1), partial(power_func, 2)])
        >>> # define a 3 x 2 input
        >>> inp = np.arange(1, 7).reshape(3, 2)
        >>> X = bas.compute_features(inp)
        >>> X.shape  # (3, 2 * 2)
        (3, 4)
        >>> bas.split_by_feature(X)["CustomBasis"]  # spilt to (3, 2, 2)
        array([[[ 1.,  1.],
                [ 2.,  4.]],
        ...
               [[ 3.,  9.],
                [ 4., 16.]],
        ...
               [[ 5., 25.],
                [ 6., 36.]]])
        """
        # ruff: noqa: D205, D400
        return super().split_by_feature(x, axis=axis)

    def set_input_shape(self, *xi: "int | tuple[int, ...] | NDArray"):
        """
        Set the expected input shape for the basis object.

        This method sets the input shape for each input required by the funcs in the CustomBasis.
        One ``xi`` must be provided for each input, specified as an integer,
        a tuple of integers, or an array. The method calculates and stores the total number of output features
        based on the number of basis functions, the number of input per function, and the provided input shapes.

        Parameters
        ----------
        *xi :
            The input shape specifications. For every k,``xi[k]`` can be:
            - An integer: Represents the dimensionality of the input. A value of ``1`` is treated as scalar input.
            - A tuple: Represents the exact input shape excluding the first axis (sample axis).
              All elements must be integers.
            - An array: The shape is extracted, excluding the first axis (assumed to be the sample axis).

        Raises
        ------
        ValueError
            If a tuple is provided, and it contains non-integer elements.
            If not enough inputs are provided.

        Returns
        -------
        self :
            Returns the instance itself to allow method chaining.

        Examples
        --------
        >>> import nemos as nmo
        >>> import numpy as np
        >>> from functools import partial
        >>> # Basis with one input only
        >>> def power_func(n, x):
        ...     return x ** n
        >>> basis = nmo.basis.CustomBasis([partial(power_func, n) for n in range(1, 6)])
        >>> # Configure with an integer input:
        >>> _ = basis.set_input_shape(3)
        >>> basis.n_output_features
        15
        >>> # Configure with a tuple:
        >>> _ = basis.set_input_shape((4, 5))
        >>> basis.n_output_features
        100
        >>> # Configure with an array:
        >>> x = np.ones((10, 4, 5))
        >>> _ = basis.set_input_shape(x)
        >>> basis.n_output_features
        100
        >>> # basis with 2 inputs
        >>> def power_add_func(n, x, y):
        ...     return x ** n + y ** n
        >>> basis = nmo.basis.CustomBasis([partial(power_add_func, n) for n in range(1, 6)])
        >>> _ = basis.set_input_shape(3, 3)
        >>> basis.n_output_features
        15
        >>> _ = basis.set_input_shape((3, 2), (3, 2))
        >>> basis.n_output_features
        30
        >>> _ = basis.set_input_shape(np.ones((10, 3, 2)), (3, 2))
        >>> basis.n_output_features
        30
        """
        super().set_input_shape(*xi, allow_inputs_of_different_shape=False)
        # CustomBasis acts as a multiplicative basis in n-dimension
        # i.e. multiple inputs must have the same shape and are
        # treated in a paired-way in vectorization
        self._input_shape_ = (
            None if self._input_shape_ is None else self._input_shape_[:1]
        )
        return self

    def to_transformer(self) -> "TransformerBasis":
        """
        Turn the Basis into a TransformerBasis for use with scikit-learn.

        Returns
        -------
        :
            A transformer basis.

        Examples
        --------
        >>> from functools import partial
        >>>
        >>> import numpy as np
        >>> from sklearn.model_selection import GridSearchCV
        >>> from sklearn.pipeline import Pipeline
        >>>
        >>> import nemos as nmo
        >>>
        >>> # load some data
        >>> x = 0.1 * np.random.normal(size=(100, 1))
        >>> y = np.random.poisson(np.exp(x[:, 0]), size=100)
        >>>
        >>>
        >>> def power_func(n, x, bias=0):
        ...     return (x + bias) ** n
        >>>
        >>>
        >>> basis = nmo.basis.CustomBasis([partial(power_func, n) for n in range(1, 6)])
        >>> basis = basis.to_transformer()
        >>> glm = nmo.glm.GLM(regularizer="Ridge", regularizer_strength=1.0)
        >>> pipeline = Pipeline([("basis", basis), ("glm", glm)])
        >>> param_grid = dict(
        ...     glm__regularizer_strength=(0.1, 0.01, 0.001, 1e-6),
        ...     basis__basis_kwargs=(dict(bias=0), dict(bias=1)),
        ... )
        >>> gridsearch = GridSearchCV(
        ...     pipeline,
        ...     param_grid=param_grid,
        ...     cv=2,
        ... )
        >>> gridsearch = gridsearch.fit(x, y)
        """
        return super().to_transformer()

    @property
    def input_shape(
        self,
    ) -> None | List[None] | Tuple[int, ...] | List[Tuple[int, ...]]:
        """Input shape as a tuple or list of tuple.

        The property mimics the behavior of atomic bases, and uses the
        assumption that _input_shape_ for custom bases is a list of
        length one.
        """
        input_shape = self._input_shape_
        if input_shape is None:
            if self._n_input_dimensionality == 1:
                return None
            else:
                return [None] * self._n_input_dimensionality
        if self._n_input_dimensionality == 1:
            return input_shape[0]
        return input_shape * self._n_input_dimensionality



from contextlib import nullcontext as does_not_raise

import jax.numpy as jnp
import pytest

from nemos.solvers import _svrg_defaults


@pytest.fixture
def x_sample():
    return jnp.array([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]])


@pytest.fixture
def y_sample():
    return jnp.array([1.0, 2.0, 3.0])


def test_convert_to_float_decorator():
    @_svrg_defaults._convert_to_float
    def sample_function(x):
        return x

    result = sample_function(1)
    assert isinstance(result, float)


def test_svrg_optimal_batch_and_stepsize(x_sample, y_sample):
    """Test calculation of optimal batch size and step size for SVRG."""
    result = _svrg_defaults.svrg_optimal_batch_and_stepsize(
        _svrg_defaults.glm_softplus_poisson_l_max_and_l,
        x_sample,
        y_sample,
        strong_convexity=0.1,
    )
    assert "batch_size" in result
    assert "stepsize" in result
    assert result["batch_size"] > 0
    assert result["stepsize"] > 0
    assert isinstance(result["batch_size"], int)
    assert isinstance(result["stepsize"], float)


@pytest.mark.parametrize("batch_size", [1, 2, 3])
def test_softplus_poisson_l_smooth_multiply(x_sample, y_sample, batch_size):
    """Test multiplication with X.T @ D @ X without forming the matrix."""
    v_sample = jnp.array([0.5, 1.0])
    result = _svrg_defaults._glm_softplus_poisson_l_smooth_multiply(
        x_sample, y_sample, v_sample, batch_size
    )
    diag_mat = jnp.diag(y_sample * 0.17 + 0.25)
    expected_result = (
        x_sample.T.dot(diag_mat).dot(x_sample.dot(v_sample)) / x_sample.shape[0]
    )
    assert jnp.allclose(result, expected_result)


def test_softplus_poisson_l_smooth_with_power_iteration(x_sample, y_sample):
    """Test the power iteration method for finding the largest eigenvalue."""
    result = _svrg_defaults._glm_softplus_poisson_l_smooth_with_power_iteration(
        x_sample, y_sample, n_power_iters=20, batch_size=x_sample.shape[0]
    )
    # compute eigvals directly
    diag_mat = jnp.diag(y_sample * 0.17 + 0.25)
    XDX = x_sample.T.dot(diag_mat).dot(x_sample) / x_sample.shape[0]
    eigmax = jnp.linalg.eigvalsh(XDX).max()

    assert result > 0
    assert jnp.allclose(eigmax, result)


@pytest.mark.parametrize("batch_size", [1, 2, 10])
@pytest.mark.parametrize("num_samples", [10, 12, 100, 500])
@pytest.mark.parametrize("l_smooth_max", [0.1, 1.0, 10.0])
@pytest.mark.parametrize("l_smooth", [0.01, 0.05, 2.0])
def test_calculate_stepsize_svrg(batch_size, num_samples, l_smooth_max, l_smooth):
    """Test calculation of the optimal step size for SVRG."""
    stepsize = _svrg_defaults._calculate_stepsize_svrg(
        batch_size, num_samples, l_smooth_max, l_smooth
    )
    assert stepsize > 0
    assert isinstance(stepsize, float)


@pytest.mark.parametrize("num_samples", [12, 100, 500])
@pytest.mark.parametrize("l_smooth_max", [0.1, 1.0, 10.0])
@pytest.mark.parametrize("l_smooth", [0.01, 0.05])
@pytest.mark.parametrize("strong_convexity", [0.01, 1.0, 10.0])
def test_calculate_optimal_batch_size_svrg(
    num_samples, l_smooth_max, l_smooth, strong_convexity
):
    """Test calculation of the optimal batch size for SVRG."""
    batch_size = _svrg_defaults._calculate_optimal_batch_size_svrg(
        num_samples, l_smooth_max, l_smooth, strong_convexity
    )
    assert batch_size > 0
    assert isinstance(batch_size, int)


@pytest.mark.parametrize(
    "num_samples, l_smooth_max, l_smooth, expected_b_hat",
    [
        (100, 10.0, 2.0, 2.8697202),
        (121, 11.0, 1.0, 4.690416),
    ],
)
def test_calculate_b_hat(num_samples, l_smooth_max, l_smooth, expected_b_hat):
    """Test calculation of b_hat for SVRG."""
    b_hat = _svrg_defaults._calculate_b_hat(num_samples, l_smooth_max, l_smooth)
    assert jnp.isclose(b_hat, expected_b_hat)


@pytest.mark.parametrize(
    "num_samples, l_smooth_max, l_smooth, strong_convexity, expected_b_tilde",
    [
        (100, 10.0, 2.0, 0.1, 3.4146341463414633),
        (121, 11.0, 1.0, 0.4, 0.6769230769230770),
    ],
)
def test_calculate_b_tilde(
    num_samples, l_smooth_max, l_smooth, strong_convexity, expected_b_tilde
):
    """Test calculation of b_tilde for SVRG."""
    b_tilde = _svrg_defaults._calculate_b_tilde(
        num_samples, l_smooth_max, l_smooth, strong_convexity
    )
    assert jnp.isclose(b_tilde, expected_b_tilde)


@pytest.mark.parametrize(
    "batch_size, stepsize, expected_batch_size, expected_stepsize",
    [
        (32, 0.01, 32, 0.01),  # Both batch_size and stepsize provided
        (32, None, 32, None),  # Only batch_size provided
        (None, 0.01, None, 0.01),  # Only stepsize provided
    ],
)
def test_svrg_optimal_batch_and_stepsize_with_provided_defaults(
    batch_size, stepsize, expected_batch_size, expected_stepsize, x_sample, y_sample
):
    """Test that provided defaults for batch_size and stepsize are returned as-is or computed correctly."""
    x_sample = jnp.tile(x_sample, 33).reshape(-1, x_sample.shape[-1])
    y_sample = jnp.tile(y_sample, 33)
    result = _svrg_defaults.svrg_optimal_batch_and_stepsize(
        _svrg_defaults.glm_softplus_poisson_l_max_and_l,
        x_sample,
        y_sample,
        batch_size=batch_size,
        stepsize=stepsize,
    )
    if expected_batch_size is not None:
        assert (
            result["batch_size"] == expected_batch_size
        ), "Provided batch_size should be returned as-is."
    else:
        assert (
            "batch_size" in result and result["batch_size"] > 0
        ), "Batch size should be computed since it was not provided."
    if expected_stepsize is not None:
        assert (
            result["stepsize"] == expected_stepsize
        ), "Provided stepsize should be returned as-is."
    else:
        assert (
            "stepsize" in result and result["stepsize"] > 0
        ), "Stepsize should be computed since it was not provided."


@pytest.mark.parametrize(
    "batch_size, stepsize, strong_convexity, expectation",
    [
        (
            32,
            None,
            0.1,
            pytest.warns(
                UserWarning, match="Could not determine step size automatically"
            ),
        ),
        (None, None, 0.1, does_not_raise()),
    ],
)
def test_warnigns_svrg_optimal_batch_and_stepsize(
    batch_size, stepsize, strong_convexity, expectation, x_sample, y_sample
):
    """Test that warnings are correctly raised during SVRG optimization when appropriate."""
    with expectation:
        _svrg_defaults.svrg_optimal_batch_and_stepsize(
            _svrg_defaults.glm_softplus_poisson_l_max_and_l,
            x_sample,
            y_sample,
            batch_size=batch_size,
            stepsize=stepsize,
            strong_convexity=strong_convexity,
        )


@pytest.mark.parametrize(
    "n_power_iter, expectation",
    [
        (
            None,
            pytest.warns(UserWarning, match="Direct computation of the eigenvalues"),
        ),
        (1, does_not_raise()),
        (10, does_not_raise()),
        (
            "a",
            pytest.raises(
                TypeError, match="`n_power_iters` must be an integer or None"
            ),
        ),
        (
            0.5,
            pytest.raises(
                TypeError, match="`n_power_iters` must be an integer or None"
            ),
        ),
        (-1, pytest.raises(ValueError, match="`n_power_iters` must be positive")),
    ],
)
def test_glm_softplus_poisson_l_smooth_power_iter(
    x_sample, y_sample, n_power_iter, expectation
):
    with expectation:
        _svrg_defaults._glm_softplus_poisson_l_smooth(
            x_sample, y_sample, batch_size=1, n_power_iters=n_power_iter
        )


@pytest.mark.parametrize(
    "delta_num_sample, expectation",
    [
        (0, does_not_raise()),
        (
            1,
            pytest.raises(
                ValueError, match="Each array in data must have the same number"
            ),
        ),
    ],
)
def test_svrg_optimal_batch_and_stepsize_num_samples(
    x_sample, y_sample, delta_num_sample, expectation
):
    y_sample = y_sample[delta_num_sample:]
    with expectation:
        _svrg_defaults.svrg_optimal_batch_and_stepsize(
            _svrg_defaults.glm_softplus_poisson_l_max_and_l,
            x_sample,
            y_sample,
            batch_size=1,
            stepsize=0.1,
            strong_convexity=0.1,
        )


@pytest.mark.parametrize(
    "num_samples, l_smooth_max, l_smooth, strong_convexity, expected_batch_size",
    [
        # Case 1: strong_convexity is None
        (
            100,
            10.0,
            2.0,
            None,
            1,
        ),  # strong_convexity is None, should return batch_size = 1
        # Case 2: num_samples >= 3 * l_smooth_max / strong_convexity
        (
            100,
            10.0,
            2.0,
            0.8,
            1,
        ),  # num_samples >= 3 * l_smooth_max / strong_convexity, should return batch_size = 1
        # Case 3: num_samples > l_smooth / strong_convexity
        (
            100,
            10.0,
            2.0,
            0.1,
            2,
        ),  # num_samples > l_smooth / strong_convexity, and b_tilde is the minimum
        # Case 4: l_smooth_max < num_samples * l_smooth / 3 and b_hat < b_tilde
        (
            100,
            5.0,
            0.2,
            0.1,
            1,
        ),  # l_smooth_max < num_samples * l_smooth / 3, use minimum(b_hat, b_tilde)
        # Case 5: l_smooth_max >= num_samples * l_smooth / 3
        (
            100,
            10.0,
            0.2,
            0.01,
            27,
        ),  # l_smooth_max >= num_samples * l_smooth / 3, batch_size = num_samples
        # Case 6: l_smooth_max >= num_samples * l_smooth / 3, but falls back to b_tilde
        (
            100,
            5.0,
            0.05,
            0.1,
            1,
        ),  # l_smooth_max >= num_samples * l_smooth / 3, but falls back to b_tilde
        # Case 7: l_smooth_max < num_samples * l_smooth / 3
        (100, 5.0, 0.5, 0.005, 4),
        # Case 8:  l_smooth_max > num_samples * l_smooth / 3
        (100, 18.0, 0.5, 0.005, 100),
    ],
)
def test_calculate_optimal_batch_size_svrg_all_config(
    num_samples, l_smooth_max, l_smooth, strong_convexity, expected_batch_size
):
    """Test the calculation of the optimal batch size for SVRG."""
    batch_size = _svrg_defaults._calculate_optimal_batch_size_svrg(
        num_samples, l_smooth_max, l_smooth, strong_convexity
    )
    assert (
        batch_size == expected_batch_size
    ), f"Expected batch_size {expected_batch_size}, got {batch_size}"



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Limited-memory Broyden method"""

from functools import partial

from typing import Any
from typing import Callable
from typing import NamedTuple
from typing import Optional
from typing import Union

from dataclasses import dataclass

import jax
import jax.numpy as jnp

from nemos.third_party.jaxopt.jaxopt._src import base
from nemos.third_party.jaxopt.jaxopt._src.backtracking_linesearch import BacktrackingLineSearch
from nemos.third_party.jaxopt.jaxopt.tree_util import tree_map
from nemos.third_party.jaxopt.jaxopt.tree_util import tree_vdot
from nemos.third_party.jaxopt.jaxopt.tree_util import tree_add_scalar_mul
from nemos.third_party.jaxopt.jaxopt.tree_util import tree_scalar_mul
from nemos.third_party.jaxopt.jaxopt.tree_util import tree_sub
from nemos.third_party.jaxopt.jaxopt.tree_util import tree_l2_norm
from nemos.third_party.jaxopt.jaxopt._src.tree_util import tree_single_dtype


def matvec(d_history, c_history, x, indices):
  # CD^T x
  # here j is the history dimension and i is the space dimension
  dtx = jnp.einsum('j..., ... -> j', d_history[indices], x)
  cdtx = jnp.einsum('j..., j -> ...', c_history[indices], dtx)
  return cdtx

def inv_jacobian_product_leaf(v: jnp.ndarray,
                              d_history: jnp.ndarray,
                              c_history: jnp.ndarray,
                              gamma: float = 1.0,
                              start: int= 0):
  # the computation of the jacobian inverse for
  # the Broyden method
  # this is done using the notations of DEQ
  history_size = len(d_history)

  indices = (start + jnp.arange(history_size)) % history_size

  v = gamma * v + matvec(d_history, c_history, v, indices)
  return v


def inv_jacobian_product(pytree: Any,
                         d_history: Any,
                         c_history: Any,
                         gamma: float = 1.0,
                         start: int = 0):
  """Product between an approximate jacobian inverse and a pytree.

  Histories are pytrees of the same structure as `pytree`.
  Leaves are arrays of shape `(history_size, ...)`, where
  `...` means the same shape as `pytree`'s leaves.

  The notation follows the scipy one.

  Args:
    pytree: pytree to multiply with.
    d_history: pytree with the same structure as `pytree`.
      Leaves contain v variables, i.e., `(x[k] - x[k-1])^T B / ((g[k] - g[k-1])^T (x[k] - x[k-1])^T B)`.
    c_history: pytree with the same structure as `pytree`.
      Leaves contain u variables, i.e., `(x[k] - x[k-1]) - B(g[k] - g[k-1])`.
    gamma: scalar to use for the initial inverse jacobian approximation,
      i.e., `gamma * I`.
    start: starting index in the circular buffer.
  """
  fun = partial(inv_jacobian_product_leaf,
                gamma=gamma,
                start=start)
  return tree_map(fun, pytree, d_history, c_history)

def inv_jacobian_rproduct(pytree: Any,
                          d_history: Any,
                          c_history: Any,
                          gamma: float = 1.0,
                          start: int = 0):
  return inv_jacobian_product(pytree, c_history, d_history, jnp.conjugate(gamma), start)


def init_history(pytree, history_size):
  fun = lambda leaf: jnp.zeros((history_size,) + leaf.shape, dtype=leaf.dtype)
  return tree_map(fun, pytree)


def update_history(history_pytree, new_pytree, last):
  fun = lambda history_array, new_value: history_array.at[last].set(new_value)
  return tree_map(fun, history_pytree, new_pytree)


class BroydenState(NamedTuple):
  """Named tuple containing state information."""
  iter_num: int
  value: float
  stepsize: float
  error: float
  d_history: Any
  c_history: Any
  gamma: jnp.ndarray
  aux: Optional[Any] = None
  failed_linesearch: bool = False

  num_fun_eval: int = 0
  num_linesearch_iter: int = 0


@dataclass(eq=False)
class Broyden(base.IterativeSolver):
  """Limited-memory Broyden solver.

  This method is a quasi-Newton approach to root finding.
  While similar to L-BFGS in spirit, it is not applied
  in the same situations: indeed, because the function
  whose root we are looking for is not necessarily a
  gradient, its Jacobian (i.e. its Hessian in the
  optimization case) is not necessarily symmetric.
  As a consequence, we cannot include symmetry in the
  secant conditions defining the updates of the Broyden
  matrices, and therefore the resulting Jacobian
  approximation is not symmetric, while it is for L-BFGS.
  Another consequence is that each Broyden update is of
  rank-1 while it is rank-2 for L-BFGS.

  Attributes:
    fun: a function of the form ``fun(x, *args, **kwargs)``.
    has_aux: whether ``fun`` outputs auxiliary data or not.
      If ``has_aux`` is False, ``fun`` is expected to be
        scalar-valued.
      If ``has_aux`` is True, then we have one of the following
        two cases.
      At each iteration of the algorithm, the auxiliary outputs are stored
        in ``state.aux``.

    maxiter: maximum number of Broyden iterations.
    tol: tolerance of the stopping criterion.

    stepsize: a stepsize to use (if <= 0, use backtracking line search),
      or a callable specifying the **positive** stepsize to use at each iteration.
    linesearch: the type of line search to use: for now only "backtracking" for backtracking
      line search is available.
    stop_if_linesearch_fails: whether to stop iterations if the line search fails.
      When True, this matches the behavior of core JAX.
    maxls: maximum number of iterations to use in the line search.
    decrease_factor: factor by which to decrease the stepsize during line search
      (default: 0.8).
    increase_factor: factor by which to increase the stepsize during line search
      (default: 1.5).
    max_stepsize: upper bound on stepsize.
    min_stepsize: lower bound on stepsize guess at start of the linesearch run.

    history_size: size of the memory to use.
    gamma: the initialization of the inverse Jacobian is going to be gamma * I.

    implicit_diff: whether to enable implicit diff or autodiff of unrolled
      iterations.
    implicit_diff_solve: the linear system solver to use.

    jit: whether to JIT-compile the optimization loop (default: True).
    unroll: whether to unroll the optimization loop (default: "auto").

    verbose: if set to True or 1 prints the information at each step of 
      the solver, if set to 2, print also the information of the linesearch.

  Reference:
    Charles G. Broyden.
    A Class of Methods for Solving Nonlinear Simultaneous Equations.
    Equation (4.5) (page 581).
  """

  fun: Callable
  has_aux: bool = False

  maxiter: int = 500
  tol: float = 1e-3

  stepsize: Union[float, Callable] = 0.0
  linesearch: str = "backtracking"
  stop_if_linesearch_fails: bool = False
  condition: str = "wolfe"
  maxls: int = 15
  decrease_factor: float = 0.8
  increase_factor: float = 1.5
  max_stepsize: float = 1.0
  # FIXME: should depend on whether float32 or float64 is used.
  min_stepsize: float = 1e-6

  history_size: int = None
  gamma: float = 1.0

  implicit_diff: bool = True
  implicit_diff_solve: Optional[Callable] = None

  jit: bool = True
  unroll: base.AutoOrBoolean = "auto"

  verbose: Union[bool, int] = False

  def _cond_fun(self, inputs):
    _, state = inputs[0]
    # We continue the optimization loop while the error tolerance is not met and,
    # either failed linesearch is disallowed or linesearch hasn't failed.
    return (state.error > self.tol) & jnp.logical_or(not self.stop_if_linesearch_fails, ~state.failed_linesearch)

  def init_state(self,
                 init_params: Any,
                 *args,
                 **kwargs) -> BroydenState:
    """Initialize the solver state.

    Args:
      init_params: pytree containing the initial parameters.
      *args: additional positional arguments to be passed to ``fun``.
      **kwargs: additional keyword arguments to be passed to ``fun``.
    Returns:
      state
    """
    if isinstance(init_params, base.OptStep):
      # `init_params` can either be a pytree or an OptStep object
      state_kwargs = dict(
        d_history=init_params.state.d_history,
        c_history=init_params.state.c_history,
        gamma=init_params.state.gamma,
        iter_num=init_params.state.iter_num,
        stepsize=init_params.state.stepsize,
      )
      init_params = init_params.params
      dtype = tree_single_dtype(init_params)
    else:
      dtype = tree_single_dtype(init_params)
      state_kwargs = dict(
        d_history=init_history(init_params, self.history_size),
        c_history=init_history(init_params, self.history_size),
        gamma=jnp.asarray(self.gamma, dtype=dtype),
        iter_num=jnp.asarray(0),
        stepsize=jnp.asarray(self.max_stepsize, dtype=dtype),
      )
    value, aux = self._value_with_aux(init_params, *args, **kwargs)
    return BroydenState(value=value,
                        error=jnp.asarray(jnp.inf),
                        **state_kwargs,
                        aux=aux,
                        failed_linesearch=jnp.asarray(False),
                        num_fun_eval=jnp.array(1, base.NUM_EVAL_DTYPE),
                        num_linesearch_iter=jnp.array(0, base.NUM_EVAL_DTYPE)
                        )

  def update(self,
             params: Any,
             state: BroydenState,
             *args,
             **kwargs) -> base.OptStep:
    """Performs one iteration of Broyden.

    Args:
      params: pytree containing the parameters.
      state: named tuple containing the solver state.
      *args: additional positional arguments to be passed to ``fun``.
      **kwargs: additional keyword arguments to be passed to ``fun``.
    Returns:
      (params, state)
    """
    if isinstance(params, base.OptStep):
      params = params.params

    start = state.iter_num % self.history_size
    value = state.value

    jac_prod_kwargs = dict(
      d_history=state.d_history,
      c_history=state.c_history,
      gamma=state.gamma,
      start=start,
    )

    jac_prod = partial(
      inv_jacobian_product,
      **jac_prod_kwargs,
    )

    jac_rprod = partial(
      inv_jacobian_rproduct,
      **jac_prod_kwargs,
    )

    product = jac_prod(pytree=value)
    descent_direction = tree_scalar_mul(-1.0, product)

    use_linesearch = not isinstance(self.stepsize, Callable) and self.stepsize <= 0
    if use_linesearch:
      if self.linesearch == "backtracking":
        # we need to build the function used for the line search
        # which is going to be the squared norm of the original function
        # as in scipy https://github.com/scipy/scipy/blob/main/scipy/optimize/_nonlin.py#L278
        # we then need to check if the gradient can be obtained with jax
        # and if not we can build it in the same fashion as scipy
        # https://github.com/scipy/scipy/blob/main/scipy/optimize/_nonlin.py#L285
        def ls_fun_with_aux(params, *args, **kwargs):
          f, aux = self._value_with_aux(params, *args, **kwargs)
          norm_squared = tree_l2_norm(f, squared=True)
          return norm_squared, (f, aux)
        # here we need a check if the function is not smooth
        ls_fun_with_aux_and_grad = jax.value_and_grad(ls_fun_with_aux, has_aux=True)
        ls = BacktrackingLineSearch(fun=ls_fun_with_aux_and_grad,
                                    value_and_grad=True,
                                    maxiter=self.maxls,
                                    decrease_factor=self.decrease_factor,
                                    max_stepsize=self.max_stepsize,
                                    condition=self.condition,
                                    jit=self.jit,
                                    unroll=self.unroll,
                                    has_aux=True,
                                    verbose=max(int(self.verbose)-1, 0),
                                    tol=1e-2)
        init_stepsize = jnp.where(state.stepsize <= self.min_stepsize,
                                  # If stepsize became too small, we restart it.
                                  self.max_stepsize,
                                  # Else, we increase a bit the previous one.
                                  state.stepsize * self.increase_factor)
        new_stepsize, ls_state = ls.run(init_stepsize,
                                        params, value, None,
                                        descent_direction,
                                        fun_args=args, fun_kwargs=kwargs)
        new_value, new_aux = ls_state.aux
        new_params = ls_state.params
        new_num_linesearch_iter = state.num_linesearch_iter + ls_state.iter_num
        new_num_fun_eval = state.num_fun_eval + ls_state.num_fun_eval
        failed_linesearch = ls_state.failed
      else:
        raise ValueError("Invalid name in 'linesearch' option.")
    else:
      # without line search
      if isinstance(self.stepsize, Callable):
        new_stepsize = self.stepsize(state.iter_num)
      else:
        new_stepsize = self.stepsize
      failed_linesearch = False

      new_params = tree_add_scalar_mul(params, new_stepsize, descent_direction)
      new_value, new_aux = self._value_with_aux(new_params, *args, **kwargs)
      new_num_fun_eval = state.num_fun_eval + 1
      new_num_linesearch_iter = state.num_linesearch_iter
    delta_x = tree_sub(new_params, params)
    v = jac_rprod(delta_x)
    delta_g = tree_sub(new_value, value)
    denom = 1 / tree_vdot(v, delta_g)
    d = tree_scalar_mul(denom, v)
    c = tree_sub(delta_x, jac_prod(delta_g))

    last = (start + self.history_size) % self.history_size
    d_history = update_history(state.d_history, d, last)
    c_history = update_history(state.c_history, c, last)

    new_state = BroydenState(iter_num=state.iter_num + 1,
                             value=new_value,
                             stepsize=jnp.asarray(new_stepsize),
                             error=tree_l2_norm(new_value),
                             d_history=d_history,
                             c_history=c_history,
                             gamma=state.gamma,
                             aux=new_aux,
                             num_fun_eval=new_num_fun_eval,
                             num_linesearch_iter=new_num_linesearch_iter,
                             failed_linesearch=failed_linesearch)

    if self.verbose:
      self.log_info(
          new_state,
          error_name="Norm Output",
          additional_info={
              "Stepsize": new_stepsize,
              "Number Linesearch Iterations": 
              new_state.num_linesearch_iter - state.num_linesearch_iter
          }
      )
    return base.OptStep(params=new_params, state=new_state)

  def optimality_fun(self, params, *args, **kwargs):
    """Optimality function mapping compatible with ``@custom_root``."""
    value = self._value_fun(params, *args, **kwargs)
    return value

  def _value_fun(self, params, *args, **kwargs):
    if isinstance(params, base.OptStep):
      params = params.params
    value, _ = self._value_with_aux(params, *args, **kwargs)
    return value

  def __post_init__(self):
    super().__post_init__()

    if self.has_aux:
      fun_ = self.fun
    else:
      fun_ = lambda p, *a, **kw: (self.fun(p, *a, **kw), None)

    self._value_with_aux = fun_

    self.reference_signature = self.fun

    if self.history_size is None:
      self.history_size = self.maxiter
